# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import json
import logging
import superdesk
import superdesk.signals as signals

from copy import copy
from copy import deepcopy
from functools import partial
from flask import current_app as app
from eve.utils import ParsedRequest

from superdesk import get_resource_service
from apps.content import push_content_notification
from apps.content_types.content_types import DEFAULT_SCHEMA
from superdesk.errors import InvalidStateTransitionError, SuperdeskApiError, SuperdeskValidationError
from superdesk.metadata.item import CONTENT_TYPE, ITEM_TYPE, ITEM_STATE, CONTENT_STATE, \
    PUBLISH_STATES, EMBARGO, PUB_STATUS, PUBLISH_SCHEDULE, SCHEDULE_SETTINGS, ASSOCIATIONS, MEDIA_TYPES
from superdesk.metadata.packages import LINKED_IN_PACKAGES, PACKAGE, PACKAGE_TYPE
from superdesk.metadata.utils import item_url
from superdesk.notification import push_notification
from superdesk.publish import SUBSCRIBER_TYPES
from superdesk.services import BaseService
from superdesk.utc import utcnow
from superdesk.workflow import is_workflow_state_transition_valid
from superdesk.validation import ValidationError

from eve.utils import config
from eve.versioning import resolve_document_version

from apps.archive.archive import ArchiveResource, SOURCE as ARCHIVE
from apps.archive.common import get_user, insert_into_versions, item_operations, \
    FIELDS_TO_COPY_FOR_ASSOCIATED_ITEM, remove_unwanted
from apps.archive.common import validate_schedule, ITEM_OPERATION, update_schedule_settings, \
    convert_task_attributes_to_objectId, get_expiry, get_utc_schedule, get_expiry_date, transtype_metadata
from apps.archive.usage import track_usage, update_refs
from apps.common.components.utils import get_component
from apps.item_autosave.components.item_autosave import ItemAutosave
from apps.legal_archive.commands import import_into_legal_archive
from apps.packages.package_service import PackageService
from apps.publish.published_item import LAST_PUBLISHED_VERSION, PUBLISHED,\
    PUBLISHED_IN_PACKAGE
from superdesk.media.crop import CropService
from superdesk.vocabularies import is_related_content
from superdesk.default_settings import strtobool

from flask_babel import _
from flask import request, json


logger = logging.getLogger(__name__)

ITEM_PUBLISH = 'publish'
ITEM_CORRECT = 'correct'
ITEM_KILL = 'kill'
ITEM_TAKEDOWN = 'takedown'
ITEM_UNPUBLISH = 'unpublish'
item_operations.extend([ITEM_PUBLISH, ITEM_CORRECT, ITEM_KILL, ITEM_TAKEDOWN, ITEM_UNPUBLISH])
publish_services = {
    ITEM_PUBLISH: 'archive_publish',
    ITEM_CORRECT: 'archive_correct',
    ITEM_KILL: 'archive_kill',
    ITEM_TAKEDOWN: 'archive_takedown',
    ITEM_UNPUBLISH: 'archive_unpublish',
}

PRESERVED_FIELDS = ['headline', 'byline', 'usageterms', 'alt_text',
                    'description_text', 'copyrightholder', 'copyrightnotice']


class BasePublishResource(ArchiveResource):
    """
    Base resource class for "publish" endpoint.
    """

    def __init__(self, endpoint_name, app, service, publish_type):
        self.endpoint_name = 'archive_%s' % publish_type
        self.resource_title = endpoint_name
        self.schema[PUBLISHED_IN_PACKAGE] = {'type': 'string'}

        self.datasource = {'source': ARCHIVE}

        self.url = 'archive/{}'.format(publish_type)
        self.item_url = item_url

        self.resource_methods = []
        self.item_methods = ['PATCH']

        self.privileges = {'PATCH': publish_type}

        super().__init__(endpoint_name, app=app, service=service)


class BasePublishService(BaseService):
    """Base service for different "publish" services."""

    publish_type = 'publish'
    published_state = 'published'
    item_operation = ITEM_PUBLISH

    non_digital = partial(filter, lambda s: s.get('subscriber_type', '') == SUBSCRIBER_TYPES.WIRE)
    digital = partial(filter, lambda s: (s.get('subscriber_type', '') in {SUBSCRIBER_TYPES.DIGITAL,
                                                                          SUBSCRIBER_TYPES.ALL}))
    package_service = PackageService()

    def on_update(self, updates, original):
        self._refresh_associated_items(original)
        self._validate(original, updates)
        self._set_updates(original, updates, updates.get(config.LAST_UPDATED, utcnow()))
        convert_task_attributes_to_objectId(updates)  # ???
        transtype_metadata(updates, original)
        self._process_publish_updates(original, updates)
        self._mark_media_item_as_used(updates, original)
        update_refs(updates, original)

    def on_updated(self, updates, original):
        original = super().find_one(req=None, _id=original[config.ID_FIELD])
        updates.update(original)

        if updates[ITEM_OPERATION] not in {ITEM_KILL, ITEM_TAKEDOWN} and \
                original.get(ITEM_TYPE) in [CONTENT_TYPE.TEXT, CONTENT_TYPE.PREFORMATTED]:
            get_resource_service('archive_broadcast').on_broadcast_master_updated(updates[ITEM_OPERATION], original)

        get_resource_service('archive_broadcast').reset_broadcast_status(updates, original)
        push_content_notification([updates])
        self._import_into_legal_archive(updates)
        CropService().update_media_references(updates, original, True)
        signals.item_published.send(self, item=original)
        packages = self.package_service.get_packages(original[config.ID_FIELD])
        if packages and packages.count() > 0:
            archive_correct = get_resource_service('archive_correct')
            processed_packages = []
            for package in packages:
                original_updates = {'operation': updates['operation'], ITEM_STATE: updates[ITEM_STATE]}
                if package[ITEM_STATE] in [CONTENT_STATE.PUBLISHED, CONTENT_STATE.CORRECTED] and \
                        package.get(PACKAGE_TYPE, '') == '' and \
                        str(package[config.ID_FIELD]) not in processed_packages:
                    original_updates['groups'] = package['groups']

                    if updates.get('headline'):
                        self.package_service.update_field_in_package(original_updates, original[config.ID_FIELD],
                                                                     'headline', updates.get('headline'))

                    if updates.get('slugline'):
                        self.package_service.update_field_in_package(original_updates, original[config.ID_FIELD],
                                                                     'slugline', updates.get('slugline'))

                    archive_correct.patch(id=package[config.ID_FIELD], updates=original_updates)
                    insert_into_versions(id_=package[config.ID_FIELD])
                    processed_packages.append(package[config.ID_FIELD])

    def update(self, id, updates, original):
        """
        Handles workflow of each Publish, Corrected, Killed and TakeDown.
        """
        try:
            user = get_user()
            auto_publish = updates.get('auto_publish', False)

            if original[ITEM_TYPE] == CONTENT_TYPE.COMPOSITE:
                self._publish_package_items(original, updates)
                self._update_archive(original, updates, should_insert_into_versions=auto_publish)
            else:
                self._publish_associated_items(original, updates)
                updated = deepcopy(original)
                updated.update(deepcopy(updates))

                if updates.get(ASSOCIATIONS):
                    self._refresh_associated_items(updated, skip_related=True)  # updates got lost with update

                if updated.get(ASSOCIATIONS):
                    self._fix_related_references(updated, updates)

                signals.item_publish.send(self, item=updated)
                self._update_archive(original, updates, should_insert_into_versions=auto_publish)
                self.update_published_collection(published_item_id=original[config.ID_FIELD], updated=updated)

            from apps.publish.enqueue import enqueue_published
            enqueue_published.apply_async()

            push_notification('item:publish', item=str(id),
                              unique_name=original['unique_name'],
                              desk=str(original.get('task', {}).get('desk', '')),
                              user=str(user.get(config.ID_FIELD, '')))

            if updates.get('previous_marked_user') and not updates.get('marked_for_user'):
                # send notification so that marked for me list can be updated
                get_resource_service('archive').handle_mark_user_notifications(updates, original, False)

        except SuperdeskApiError:
            raise
        except KeyError as e:
            logger.exception(e)
            raise SuperdeskApiError.badRequestError(
                message=_("Key is missing on article to be published: {exception}").format(exception=str(e))
            )
        except Exception as e:
            logger.exception(e)
            raise SuperdeskApiError.internalError(
                message=_("Failed to publish the item: {id}").format(id=str(id)), exception=e)

    def is_targeted(self, article, target=None):
        """Checks if article is targeted.

        Returns True if the given article has been targeted by region or
        subscriber type or specific subscribers.

        :param article: Article to check
        :param target: Optional specific target to check if exists
        :return:
        """
        if target:
            return len(article.get(target, [])) > 0
        else:
            return len(article.get('target_regions', []) +
                       article.get('target_types', []) +
                       article.get('target_subscribers', [])) > 0

    def _validate(self, original, updates):
        self.raise_if_invalid_state_transition(original)
        self._raise_if_unpublished_related_items(original)

        updated = original.copy()
        updated.update(updates)

        self.raise_if_not_marked_for_publication(updated)

        if self.publish_type == 'publish':
            # The publish schedule has not been cleared
            if updates.get(PUBLISH_SCHEDULE) or updated.get(SCHEDULE_SETTINGS, {}).get(
                    'utc_{}'.format(PUBLISH_SCHEDULE)) or not original.get(PUBLISH_SCHEDULE):
                update_schedule_settings(updated, PUBLISH_SCHEDULE, updated.get(PUBLISH_SCHEDULE))
                validate_schedule(updated.get(SCHEDULE_SETTINGS, {}).get('utc_{}'.format(PUBLISH_SCHEDULE)))

        if original[ITEM_TYPE] != CONTENT_TYPE.COMPOSITE and updates.get(EMBARGO):
            update_schedule_settings(updated, EMBARGO, updated.get(EMBARGO))
            get_resource_service(ARCHIVE).validate_embargo(updated)

        if self.publish_type in [ITEM_CORRECT, ITEM_KILL]:
            if updates.get(EMBARGO) and not original.get(EMBARGO):
                raise SuperdeskApiError.badRequestError(_("Embargo can't be set after publishing"))

        if self.publish_type == ITEM_KILL:
            if updates.get('dateline'):
                raise SuperdeskApiError.badRequestError(_("Dateline can't be modified on kill or take down"))

        if self.publish_type == ITEM_PUBLISH and updated.get('rewritten_by'):
            rewritten_by = get_resource_service(ARCHIVE).find_one(req=None, _id=updated.get('rewritten_by'))
            if rewritten_by and rewritten_by.get(ITEM_STATE) in PUBLISH_STATES:
                raise SuperdeskApiError.badRequestError(_("Cannot publish the story after Update is published.!"))

        if self.publish_type == ITEM_PUBLISH and updated.get('rewrite_of'):
            rewrite_of = get_resource_service(ARCHIVE).find_one(req=None, _id=updated.get('rewrite_of'))
            if rewrite_of and rewrite_of.get(ITEM_STATE) not in PUBLISH_STATES:
                raise SuperdeskApiError.badRequestError(_("Can't publish update until original story is published.!"))

        publish_type = 'auto_publish' if updates.get('auto_publish') else self.publish_type
        validate_item = {'act': publish_type, 'type': original['type'], 'validate': updated}
        validation_errors = get_resource_service('validate').post([validate_item], fields=True)
        for errors, fields in validation_errors:
            if errors:
                raise SuperdeskValidationError(errors, fields)

        validation_errors = []
        self._validate_associated_items(original, updates, validation_errors)

        if original[ITEM_TYPE] == CONTENT_TYPE.COMPOSITE:
            self._validate_package(original, updates, validation_errors)

        if len(validation_errors) > 0:
            raise ValidationError(validation_errors)

    def _raise_if_unpublished_related_items(self, original):
        if not request:
            return

        if (config.PUBLISH_ASSOCIATED_ITEMS
                or not original.get(ASSOCIATIONS)
                or self.publish_type not in [ITEM_PUBLISH, ITEM_CORRECT]):
            return

        archive_service = get_resource_service('archive')
        publishing_warnings_confirmed = strtobool(request.args.get('publishing_warnings_confirmed') or 'False')

        if not publishing_warnings_confirmed:
            for key, associated_item in original.get(ASSOCIATIONS).items():
                if associated_item and is_related_content(key):
                    item = archive_service.find_one(req=None, _id=associated_item.get('_id'))
                    item = item if item else associated_item

                    if item.get('state') not in PUBLISH_STATES:
                        error_msg = json.dumps({
                            'warnings': [_('There are unpublished related ' +
                                           'items that won\'t be sent out as ' +
                                           'related items. Do you want to publish the article anyway?')]
                        })
                        raise ValidationError(error_msg)

    def _validate_package(self, package, updates, validation_errors):
        # make sure package is not scheduled or spiked
        if package[ITEM_STATE] in (CONTENT_STATE.SPIKED, CONTENT_STATE.SCHEDULED):
            validation_errors.append('Package cannot be {}'.format(package[ITEM_STATE]))

        if package.get(EMBARGO):
            validation_errors.append('Package cannot have Embargo')

        items = self.package_service.get_residrefs(package)
        if self.publish_type in [ITEM_CORRECT, ITEM_KILL]:
            removed_items, added_items = self._get_changed_items(items, updates)
            # we raise error if correction is done on a empty package. Kill is fine.
            if len(removed_items) == len(items) and len(added_items) == 0 and self.publish_type == ITEM_CORRECT:
                validation_errors.append(_("Corrected package cannot be empty!"))

    def raise_if_not_marked_for_publication(self, original):
        if original.get('flags', {}).get('marked_for_not_publication', False):
            raise SuperdeskApiError.badRequestError(_('Cannot publish an item which is marked as Not for Publication'))

    def raise_if_invalid_state_transition(self, original):
        if not is_workflow_state_transition_valid(self.publish_type, original[ITEM_STATE]):
            error_message = _("Can't {operation} as item state is {state}") \
                if original[ITEM_TYPE] == CONTENT_TYPE.TEXT else \
                _("Can't {operation} as either package state or one of the items state is {state}")
            raise InvalidStateTransitionError(
                error_message.format(operation=self.publish_type, state=original[ITEM_STATE]))

    def _process_publish_updates(self, original, updates):
        """Common updates for published items."""
        desk = None
        if original.get('task', {}).get('desk'):
            desk = get_resource_service('desks').find_one(req=None, _id=original['task']['desk'])
        if not original.get('ingest_provider'):
            updates['source'] = desk['source'] if desk and desk.get('source', '') \
                else app.settings['DEFAULT_SOURCE_VALUE_FOR_MANUAL_ARTICLES']
        updates['pubstatus'] = PUB_STATUS.CANCELED if self.publish_type == ITEM_KILL else PUB_STATUS.USABLE
        self._set_item_expiry(updates, original)

    def _set_item_expiry(self, updates, original):
        """Set the expiry for the item.

        :param dict updates: doc on which publishing action is performed
        """
        desk_id = original.get('task', {}).get('desk')
        stage_id = original.get('task', {}).get('stage')

        if EMBARGO in updates or PUBLISH_SCHEDULE in updates:
            offset = get_utc_schedule(updates, PUBLISH_SCHEDULE) or get_utc_schedule(updates, EMBARGO)
        elif EMBARGO in original or PUBLISH_SCHEDULE in original:
            offset = get_utc_schedule(original, PUBLISH_SCHEDULE) or get_utc_schedule(original, EMBARGO)

        if app.settings.get('PUBLISHED_CONTENT_EXPIRY_MINUTES'):
            updates['expiry'] = get_expiry_date(app.settings['PUBLISHED_CONTENT_EXPIRY_MINUTES'], offset=offset)
        else:
            updates['expiry'] = get_expiry(desk_id, stage_id, offset=offset)

    def _publish_package_items(self, package, updates):
        """Publishes all items of a package recursively then publishes the package itself.

        :param package: package to publish
        :param updates: payload
        """
        items = self.package_service.get_residrefs(package)

        if len(items) == 0 and self.publish_type == ITEM_PUBLISH:
            raise SuperdeskApiError.badRequestError(_("Empty package cannot be published!"))

        removed_items = []
        if self.publish_type in [ITEM_CORRECT, ITEM_KILL]:
            removed_items, added_items = self._get_changed_items(items, updates)
            # we raise error if correction is done on a empty package. Kill is fine.
            if len(removed_items) == len(items) and len(added_items) == 0 and self.publish_type == ITEM_CORRECT:
                raise SuperdeskApiError.badRequestError(_("Corrected package cannot be empty!"))
            items.extend(added_items)

        if not updates.get('groups') and package.get('groups'):  # this saves some typing in tests
            updates['groups'] = package.get('groups')

        if items:
            archive_publish = get_resource_service('archive_publish')
            for guid in items:
                package_item = super().find_one(req=None, _id=guid)

                if not package_item:
                    raise SuperdeskApiError.badRequestError(
                        _("Package item with id: {guid} does not exist.").format(guid=guid))

                if package_item[ITEM_STATE] not in PUBLISH_STATES:  # if the item is not published then publish it
                    if package_item[ITEM_TYPE] == CONTENT_TYPE.COMPOSITE:
                        # if the item is a package do recursion to publish
                        sub_updates = {i: updates[i] for i in ['state', 'operation'] if i in updates}
                        sub_updates['groups'] = list(package_item['groups'])
                        self._publish_package_items(package_item, sub_updates)
                        self._update_archive(original=package_item, updates=sub_updates,
                                             should_insert_into_versions=False)
                    else:
                        # publish the item
                        package_item[PUBLISHED_IN_PACKAGE] = package[config.ID_FIELD]
                        archive_publish.patch(id=package_item.pop(config.ID_FIELD), updates=package_item)

                    insert_into_versions(id_=guid)

                elif guid in removed_items:
                    # remove the package information from the package item.
                    linked_in_packages = [linked for linked in package_item.get(LINKED_IN_PACKAGES, [])
                                          if linked.get(PACKAGE) != package.get(config.ID_FIELD)]
                    super().system_update(guid, {LINKED_IN_PACKAGES: linked_in_packages}, package_item)

                package_item = super().find_one(req=None, _id=guid)

                self.package_service.update_field_in_package(updates, package_item[config.ID_FIELD],
                                                             config.VERSION, package_item[config.VERSION])

                if package_item.get(ASSOCIATIONS):
                    self.package_service.update_field_in_package(
                        updates,
                        package_item[config.ID_FIELD],
                        ASSOCIATIONS,
                        package_item[ASSOCIATIONS]
                    )

        updated = deepcopy(package)
        updated.update(updates)
        self.update_published_collection(published_item_id=package[config.ID_FIELD], updated=updated)

    def update_published_collection(self, published_item_id, updated=None):
        """Updates the published collection with the published item.

        Set the last_published_version to false for previous versions of the published items.

        :param: str published_item_id: _id of the document.
        """
        published_item = super().find_one(req=None, _id=published_item_id)
        published_item = copy(published_item)
        if updated:
            published_item.update(updated)
        get_resource_service(PUBLISHED).update_published_items(published_item_id, LAST_PUBLISHED_VERSION, False)
        return get_resource_service(PUBLISHED).post([published_item])

    def set_state(self, original, updates):
        """Set the state of the document based on the action (publish, correction, kill, recalled)

        :param dict original: original document
        :param dict updates: updates related to document
        """
        updates[PUBLISH_SCHEDULE] = None
        updates[SCHEDULE_SETTINGS] = {}
        updates[ITEM_STATE] = self.published_state

    def _set_updates(self, original, updates, last_updated, preserve_state=False):
        """Sets config.VERSION, config.LAST_UPDATED, ITEM_STATE in updates document.

        If item is being published and embargo is available then append Editorial Note with 'Embargoed'.

        :param dict original: original document
        :param dict updates: updates related to the original document
        :param datetime last_updated: datetime of the updates.
        """
        if not preserve_state:
            self.set_state(original, updates)
        updates.setdefault(config.LAST_UPDATED, last_updated)

        if original[config.VERSION] == updates.get(config.VERSION, original[config.VERSION]):
            resolve_document_version(document=updates, resource=ARCHIVE, method='PATCH', latest_doc=original)

        user = get_user()
        if user and user.get(config.ID_FIELD):
            updates['version_creator'] = user[config.ID_FIELD]

    def _update_archive(self, original, updates, versioned_doc=None, should_insert_into_versions=True):
        """Updates the articles into archive collection and inserts the latest into archive_versions.

        Also clears autosaved versions if any.

        :param: versioned_doc: doc which can be inserted into archive_versions
        :param: should_insert_into_versions if True inserts the latest document into versions collection
        """
        self.backend.update(self.datasource, original[config.ID_FIELD], updates, original)
        app.on_archive_item_updated(updates, original, updates[ITEM_OPERATION])

        if should_insert_into_versions:
            if versioned_doc is None:
                insert_into_versions(id_=original[config.ID_FIELD])
            else:
                insert_into_versions(doc=versioned_doc)

        get_component(ItemAutosave).clear(original[config.ID_FIELD])

    def _get_changed_items(self, existing_items, updates):
        """Returns the added and removed items from existing_items.

        :param existing_items: Existing list
        :param updates: Changes
        :return: list of removed items and list of added items
        """
        if 'groups' in updates:
            new_items = self.package_service.get_residrefs(updates)
            removed_items = list(set(existing_items) - set(new_items))
            added_items = list(set(new_items) - set(existing_items))
            return removed_items, added_items
        else:
            return [], []

    def _validate_associated_items(self, original_item, updates=None, validation_errors=None):
        """Validates associated items.

        This function will ensure that the unpublished content validates and none of
        the content is locked, also do not allow any killed or recalled or spiked content.

        :param package:
        :param validation_errors: validation errors are appended if there are any.
        """

        if validation_errors is None:
            validation_errors = []

        if updates is None:
            updates = {}

        # merge associations
        associations = deepcopy(original_item.get(ASSOCIATIONS, {}))
        associations.update(updates.get(ASSOCIATIONS, {}))

        items = [value for value in associations.values()]
        if original_item[ITEM_TYPE] == CONTENT_TYPE.COMPOSITE and \
                self.publish_type == ITEM_PUBLISH:
            items.extend(self.package_service.get_residrefs(original_item))

        for item in items:
            orig = None
            if type(item) == dict and item.get(config.ID_FIELD):
                doc = item
                orig = super().find_one(req=None, _id=item[config.ID_FIELD])
                try:
                    doc.update({'lock_user': orig['lock_user']})
                except (TypeError, KeyError):
                    pass
            elif item:
                doc = super().find_one(req=None, _id=item)
            else:
                continue

            if not doc:
                continue

            if not orig:
                orig = doc.copy()

            if original_item[ITEM_TYPE] == CONTENT_TYPE.COMPOSITE:
                self._validate_associated_items(doc, validation_errors=validation_errors)

            # make sure no items are killed or recalled or spiked or scheduled
            # using the latest version of the item from archive
            doc_item_state = orig.get(ITEM_STATE, CONTENT_STATE.PUBLISHED)
            if doc_item_state in {CONTENT_STATE.KILLED, CONTENT_STATE.RECALLED,
                                  CONTENT_STATE.SPIKED, CONTENT_STATE.SCHEDULED}:
                validation_errors.append('Item cannot contain associated {} item'.format(doc[ITEM_STATE]))

            if doc.get(EMBARGO):
                validation_errors.append('Item cannot have associated items with Embargo')

            # don't validate items that already have published
            if doc_item_state not in [CONTENT_STATE.PUBLISHED, CONTENT_STATE.CORRECTED]:
                validate_item = {'act': self.publish_type, 'type': doc[ITEM_TYPE], 'validate': doc}
                if type(item) == dict:
                    validate_item['embedded'] = True
                errors = get_resource_service('validate').post([validate_item], headline=True, fields=True)[0]
                if errors[0]:
                    pre_errors = ['Associated item %s %s' % (doc.get('slugline', ''), error) for error in errors[0]]
                    validation_errors.extend(pre_errors)

            if config.PUBLISH_ASSOCIATED_ITEMS:
                # check the locks on the items
                if doc.get('lock_user'):
                    if original_item['lock_user'] != doc['lock_user']:
                        validation_errors.extend([
                            '{}: {}'.format(
                                doc.get('headline', doc['_id']),
                                _('packaged item is locked by another user')
                            )
                        ])
                    elif original_item['lock_user'] == doc['lock_user']:
                        validation_errors.extend([
                            '{}: {}'.format(
                                doc.get('headline', doc['_id']),
                                _('packaged item is locked by you. Unlock it and try again')
                            )
                        ])

    def _import_into_legal_archive(self, doc):
        """Import into legal archive async

        :param {dict} doc: document to be imported
        """

        if doc.get(ITEM_STATE) != CONTENT_STATE.SCHEDULED:
            kwargs = {
                'item_id': doc.get(config.ID_FIELD)
            }
            # countdown=3 is for elasticsearch to be refreshed with archive and published changes
            import_into_legal_archive.apply_async(countdown=3, kwargs=kwargs)  # @UndefinedVariable

    def _refresh_associated_items(self, original, skip_related=False):
        """Refreshes associated items with the latest version. Any further updates made to basic metadata done after
        item was associated will be carried on and used when validating those items.
        """
        associations = original.get(ASSOCIATIONS) or {}
        for name, item in associations.items():
            if type(item) == dict and item.get(config.ID_FIELD) and (not skip_related or len(item.keys()) > 2):
                keys = [key for key in DEFAULT_SCHEMA.keys() if key not in PRESERVED_FIELDS]

                if app.settings.get('COPY_METADATA_FROM_PARENT') and item.get(ITEM_TYPE) in MEDIA_TYPES:
                    updates = original
                    keys = FIELDS_TO_COPY_FOR_ASSOCIATED_ITEM
                else:
                    updates = super().find_one(req=None, _id=item[config.ID_FIELD]) or {}

                try:
                    is_db_item_bigger_ver = updates['_current_version'] > item['_current_version']
                except KeyError:
                    update_item_data(item, updates, keys)
                else:
                    # if copying from parent the don't keep the existing
                    # otherwise check the value is_db_item_bigger_ver
                    keep_existing = not app.settings.get('COPY_METADATA_FROM_PARENT') and not is_db_item_bigger_ver
                    update_item_data(item, updates, keys, keep_existing=keep_existing)

    def _fix_related_references(self, updated, updates):
        for key, item in updated[ASSOCIATIONS].items():
            if item and item.get('_fetchable', True) and is_related_content(key):
                updated[ASSOCIATIONS][key] = {
                    '_id': item['_id'],
                    'type': item['type'],
                    'order': item.get('order', 1),
                }
                updates.setdefault('associations', {})[key] = updated[ASSOCIATIONS][key]

    def _publish_associated_items(self, original, updates=None):
        """If there any updates to associated item and if setting:PUBLISH_ASSOCIATED_ITEMS is true
        then publish the associated item
        """

        if updates is None:
            updates = {}

        if not publish_services.get(self.publish_type):
            # publish type not supported
            return

        publish_service = get_resource_service(publish_services.get(self.publish_type))

        if not updates.get(ASSOCIATIONS) and not original.get(ASSOCIATIONS):
            # there's nothing to update
            return

        associations = original.get(ASSOCIATIONS) or {}

        if updates and updates.get(ASSOCIATIONS):
            associations.update(updates[ASSOCIATIONS])

        archive_service = get_resource_service('archive')

        for associations_key, associated_item in associations.items():
            if associated_item is None:
                continue

            if type(associated_item) == dict and associated_item.get(config.ID_FIELD):
                if not config.PUBLISH_ASSOCIATED_ITEMS or not publish_service:
                    if original.get(ASSOCIATIONS, {}).get(associations_key):
                        # Not allowed to publish
                        original[ASSOCIATIONS][associations_key]['state'] = self.published_state
                        original[ASSOCIATIONS][associations_key]['operation'] = self.publish_type
                    continue

                # if item is not fetchable, only mark it as published
                if not associated_item.get('_fetchable', True):
                    associated_item['state'] = self.published_state
                    associated_item['operation'] = self.publish_type
                    updates[ASSOCIATIONS] = updates.get(ASSOCIATIONS, {})
                    updates[ASSOCIATIONS][associations_key] = associated_item
                    continue

                if associated_item.get('state') == CONTENT_STATE.UNPUBLISHED:
                    # get the original associated item from archive
                    orig_associated_item = archive_service.find_one(req=None, _id=associated_item[config.ID_FIELD])

                    orig_associated_item['state'] = updates.get('state', self.published_state)
                    orig_associated_item['operation'] = self.publish_type

                    # if main item is scheduled we must also schedule associations
                    self._inherit_publish_schedule(original, updates, orig_associated_item)

                    get_resource_service('archive_publish').patch(id=orig_associated_item.pop(config.ID_FIELD),
                                                                  updates=orig_associated_item)
                    continue

                if associated_item.get('state') not in PUBLISH_STATES:
                    # This associated item has not been published before
                    associated_item.get('task', {}).pop('stage', None)
                    remove_unwanted(associated_item)

                    # get the original associated item from archive
                    orig_associated_item = archive_service.find_one(req=None, _id=associated_item[config.ID_FIELD])

                    # check if the original associated item exists in archive
                    if not orig_associated_item:
                        raise SuperdeskApiError.badRequestError(
                            _('Associated item "{}" does not exist in the system'.format(associations_key)))

                    if orig_associated_item.get('state') in PUBLISH_STATES:
                        # item was published already
                        original[ASSOCIATIONS][associations_key].update({
                            'state': orig_associated_item['state'],
                            'operation': orig_associated_item.get('operation', self.publish_type),
                        })
                        continue

                    # update _updated, otherwise it's stored as string.
                    # fixes SDESK-5043
                    associated_item['_updated'] = utcnow()

                    # if main item is scheduled we must also schedule associations
                    self._inherit_publish_schedule(original, updates, associated_item)

                    get_resource_service('archive_publish').patch(id=associated_item.pop(config.ID_FIELD),
                                                                  updates=associated_item)
                    associated_item['state'] = updates.get('state', self.published_state)
                    associated_item['operation'] = self.publish_type
                    updates[ASSOCIATIONS] = updates.get(ASSOCIATIONS, {})
                    updates[ASSOCIATIONS][associations_key] = associated_item
                elif associated_item.get('state') != self.published_state:
                    # Check if there are updates to associated item
                    association_updates = updates.get(ASSOCIATIONS, {}).get(associations_key)

                    # if main item is scheduled we must also schedule associations
                    self._inherit_publish_schedule(original, updates, associated_item)

                    if not association_updates:
                        # there is no update for this item
                        associated_item.get('task', {}).pop('stage', None)
                        remove_unwanted(associated_item)
                        publish_service.patch(id=associated_item.pop(config.ID_FIELD), updates=associated_item)
                        continue

                    if association_updates.get('state') not in PUBLISH_STATES:
                        # There's an update to the published associated item
                        remove_unwanted(association_updates)
                        publish_service.patch(id=association_updates.pop(config.ID_FIELD), updates=association_updates)

        self._refresh_associated_items(original)

    def _mark_media_item_as_used(self, updates, original):
        if ASSOCIATIONS not in updates or not updates.get(ASSOCIATIONS):
            return

        for item_name, item_obj in updates.get(ASSOCIATIONS).items():
            if not item_obj or config.ID_FIELD not in item_obj:
                continue
            item_id = item_obj[config.ID_FIELD]
            media_item = self.find_one(req=None, _id=item_id)
            if app.settings.get('COPY_METADATA_FROM_PARENT') and item_obj.get(ITEM_TYPE) in MEDIA_TYPES:
                stored_item = (original.get(ASSOCIATIONS) or {}).get(item_name) or item_obj
            else:
                stored_item = media_item
                if not stored_item:
                    continue
            track_usage(media_item, stored_item, item_obj, item_name, original)

    def _inherit_publish_schedule(self, original, updates, associated_item):
        if self.publish_type == 'publish' and (updates.get(PUBLISH_SCHEDULE) or original.get(PUBLISH_SCHEDULE)):
            schedule_settings = updates.get(SCHEDULE_SETTINGS, original.get(SCHEDULE_SETTINGS, {}))
            publish_schedule = updates.get(PUBLISH_SCHEDULE, original.get(PUBLISH_SCHEDULE))

            associated_item[PUBLISH_SCHEDULE] = publish_schedule
            associated_item[SCHEDULE_SETTINGS] = schedule_settings


def get_crop(rendition):
    fields = ('CropLeft', 'CropTop', 'CropRight', 'CropBottom')
    return {field: rendition[field] for field in fields if field in rendition}


def update_item_data(item, data, keys=None, keep_existing=False):
    """Update main item data, so only keys from default schema.

    :param dict item: item to update
    :param dict data: update date
    :param list keys: keys of item to update
    :param bool keep_existing: if True, will only set non existing values
    """
    if keys is None:
        keys = DEFAULT_SCHEMA.keys()

    for key in keys:
        if data.get(key):
            if keep_existing:
                item.setdefault(key, data[key])
            else:
                item[key] = data[key]


superdesk.workflow_state('published')
superdesk.workflow_action(
    name='publish',
    include_states=['fetched', 'routed', 'submitted', 'in_progress', 'scheduled', 'unpublished'],
    privileges=['publish']
)

superdesk.workflow_state('scheduled')
superdesk.workflow_action(
    name='schedule',
    include_states=['fetched', 'routed', 'submitted', 'in_progress'],
    privileges=['schedule']
)

superdesk.workflow_action(
    name='deschedule',
    include_states=['scheduled'],
    privileges=['deschedule']
)

superdesk.workflow_state('killed')
superdesk.workflow_action(
    name='kill',
    include_states=['published', 'scheduled', 'corrected'],
    privileges=['kill']
)

superdesk.workflow_state('corrected')
superdesk.workflow_action(
    name='correct',
    include_states=['published', 'corrected'],
    privileges=['correct']
)

superdesk.workflow_action(
    name='rewrite',
    exclude_states=['killed', 'spiked', 'scheduled'],
    privileges=['rewrite']
)

superdesk.workflow_state('recalled')
superdesk.workflow_action(
    name='recalled',
    include_states=['published', 'scheduled', 'corrected'],
    privileges=['takedown']
)

superdesk.workflow_state('unpublished')
superdesk.workflow_action(
    name='unpublish',
    include_states=['published', 'scheduled', 'corrected'],
    privileges=['unpublish']
)



import json
from flask import url_for

from superdesk import get_resource_service


def test_service_get(prodapi_app_with_data):
    """
    Test fetching items using `desks` service
    :param prodapi_app_with_data: prod api app with filled data
    """

    with prodapi_app_with_data.app_context():
        items_service = get_resource_service('desks')

        assert len(list(items_service.get(req=None, lookup={}))) == 2


def test_readonly(prodapi_app_with_data, prodapi_app_with_data_client):
    """
    Ensure that `desks` endpoint is readonly
    :param prodapi_app_with_data: prod api app with filled data
    :param prodapi_app_with_data_client: client for prod api app with filled data
    """

    with prodapi_app_with_data.test_request_context():
        for method, status in (('get', 200), ('post', 405), ('patch', 405), ('put', 405), ('delete', 405)):
            # we send a request
            resp = getattr(prodapi_app_with_data_client, method)(
                url_for('desks|resource')
            )
            # we get a response
            assert resp.status_code == status


def test_excluded_fields(prodapi_app_with_data, prodapi_app_with_data_client):
    """
    Ensure that fields which are listed as `excluded_fields` are not in the response.
    :param prodapi_app_with_data: prod api app with filled data
    :param prodapi_app_with_data_client: client for prod api app with filled data
    """

    excluded_fields = {
        'default_content_template',
        'working_stage',
        'incoming_stage',
        'desk_metadata',
        'content_expiry',
        '_etag',
        '_type',
        '_updated',
        '_created',
        '_current_version',
        '_links',
    }

    with prodapi_app_with_data.test_request_context():
        # list
        resp = prodapi_app_with_data_client.get(url_for('desks|resource'))
        resp_data = json.loads(resp.data.decode('utf-8'))

        for item in resp_data['_items']:
            assert len(set(item.keys()) & excluded_fields) == 0

        # details
        item = resp_data['_items'][0]
        resp = prodapi_app_with_data_client.get(
            url_for('desks|item_lookup', _id=item['_id']),
        )
        resp_data = json.loads(resp.data.decode('utf-8'))

        assert len(set(resp_data.keys()) & excluded_fields) == 0


def test_filters(prodapi_app_with_data, prodapi_app_with_data_client):
    """
    Test filtering.
    :param prodapi_app_with_data: prod api app with filled data
    :param prodapi_app_with_data_client: client for prod api app with filled data
    """

    with prodapi_app_with_data.test_request_context():
        params = {
            'where': json.dumps({
                'name': 'PROD'
            })
        }
        resp = prodapi_app_with_data_client.get(url_for('desks|resource', **params))
        resp_data = json.loads(resp.data.decode('utf-8'))
        assert len(resp_data['_items']) == 1
        assert resp_data['_items'][0]['_id'] == '5d76113ab3af37dea3a2eb9e'



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2019 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from superdesk.resource import Resource
from superdesk.metadata.utils import item_url
from superdesk.auth_server.scopes import Scope
from apps.contacts.resource import ContactsResource as _ContactsResource


class ContactsResource(Resource):
    url = 'contacts'
    item_url = item_url
    item_methods = ['GET']
    resource_methods = ['GET']
    schema = _ContactsResource.schema.copy()
    datasource = {
        'source': 'contacts',
        'search_backend': 'elastic',
        'default_sort': [('last_name.keyword', 1)],
        'projection': {
            'fields_meta': 0
        },
    }
    privileges = {
        'GET': Scope.CONTACTS_READ.name
    }




import bson
import superdesk

from flask import abort, request, json
from superdesk.utc import utcnow
from superdesk.utils import ListCursor
from apps.search_providers.registry import registered_search_providers
from apps.io.search_ingest import SearchIngestService
from superdesk.metadata.item import get_schema


PROXY_ENDPOINT = 'search_providers_proxy'


class SearchProviderProxyResource(superdesk.Resource):
    schema = {
        'guid': {'type': 'string', 'required': True},
        'desk': superdesk.Resource.rel('desks', False, nullable=True),
        'repo': superdesk.Resource.rel('search_providers', False, nullable=True),
        'stage': superdesk.Resource.rel('stages', False, nullable=True),
        'fetch_endpoint': {'type': 'string', 'readonly': True},
        'search_provider': {'type': 'string', 'readonly': True},
        '_fetchable': {'type': 'boolean', 'readonly': True},
    }
    schema.update(get_schema())

    resource_methods = ['GET', 'POST']
    privileges = {'POST': 'archive'}


class SearchProviderProxyService(SearchIngestService):
    """Search Provider Proxy.

    It will forward requests to specific SearchProvider instance.

    Old implementations will get forwarded to its service.
    """

    def get_provider(self, provider_id=None, req=None):
        if not provider_id:
            provider_id = req.args.get('repo') if req else request.args.get('repo')
        if provider_id is None:
            abort(400)
        try:
            bson.ObjectId(provider_id)
        except bson.errors.InvalidId:
            return provider_id
        provider = superdesk.get_resource_service('search_providers').find_one(req=None, _id=provider_id)
        if not provider:
            abort(400)
        if provider.get('is_closed'):
            abort(400)
        return provider

    def _get_service(self, provider):
        if isinstance(provider, str):
            return provider if ',' not in provider and provider else 'search'
        provider_data = registered_search_providers[provider['search_provider']]
        try:
            return provider_data['endpoint']
        except KeyError:
            return provider_data['class'](provider)

    def get(self, req, lookup):
        """Search using provider."""
        provider = self.get_provider(req=req)
        service = self._get_service(provider)
        if isinstance(service, str):
            return superdesk.get_resource_service(service).get(req, lookup)
        query = self._get_query(req)
        params = json.loads(req.args['params']) if req.args.get('params') else {}
        try:
            items = service.find(query, params)
        except TypeError:  # BC
            items = service.find(query)
        if isinstance(items, list):
            items = ListCursor(items)
        for item in items:
            self._set_item_defaults(item, provider)
        return items

    def create(self, docs, **kwargs):
        """Archive items from provider."""
        provider = self.get_provider()
        service = self._get_service(provider)
        if isinstance(service, str):
            return superdesk.get_resource_service(service).create(docs, **kwargs)
        return super().create(docs, **kwargs)

    def fetch(self, guid):
        """Fetch single item from provider to archive it."""
        provider = self.get_provider()
        service = self._get_service(provider)
        if isinstance(service, str):
            return superdesk.get_resource_service(service).fetch(guid)
        item = service.fetch(guid)
        self._set_item_defaults(item, provider)
        return item

    def fetch_rendition(self, rendition, item):
        """Fetch binary from provider."""
        provider = self.get_provider(provider_id=item.get('ingest_provider'))
        service = self._get_service(provider)
        if isinstance(service, str):
            return superdesk.get_resource_service(service).fetch_rendition(self, rendition)
        return service.fetch_file(rendition.get('href'), rendition=rendition, item=item)

    def _set_item_defaults(self, item, provider):
        """Add default values to external items."""
        now = utcnow()
        item.setdefault('_id', item.get('guid') or bson.ObjectId())
        item.setdefault('_type', 'externalsource')
        item.setdefault('type', 'picture')
        item.setdefault('pubstatus', 'usable')
        item.setdefault('firstcreated', now)
        item.setdefault('versioncreated', now)
        item.setdefault('fetch_endpoint', PROXY_ENDPOINT)
        item.setdefault('ingest_provider', str(provider['_id']))



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import re
import json
import requests

from eve.render import send_response
from flask import Blueprint

from superdesk.utc import utcnow


bp = Blueprint('vidible', __name__)


def get_vidible_metadata(bcid, pid):
    """
    Retrieve the metadata for a given vidible embed
    """
    vidible_metadata = {}
    # first we fetch a javascript file which contains some data and the video id
    embed_meta_resp = requests.get('http://delivery.vidible.tv/jsonp/pid={pid}/{bcid}.js'.format(pid=pid, bcid=bcid))
    embed_meta = re.search('({.*})', embed_meta_resp.text)
    if embed_meta:
        embed_meta = json.loads(embed_meta.group(0))
        vidible_metadata.update({
            'type': 'video',
            'bcid': embed_meta['wlcid'],
            'uri': embed_meta['bid']['id'],
            'thumbnail': embed_meta['bid']['videos'][0]['thumbnail'],
            'url': embed_meta['bid']['videos'][0]['videoUrls'][0],
            'company': embed_meta['bid']['videos'][0]['studioName'],
            'duration': embed_meta['bid']['videos'][0]['metadata']['duration'],
        })
        for video in embed_meta['bid']['videos'][:1]:
            video_id = video['videoId']
            # we use the video id in order to get additional metadata, like width, height, mimeType and more...
            video_meta_resp = requests.get('http://api.vidible.tv/search?bcid={bcid}&query={video_id}'.format(
                                           video_id=video_id, bcid=bcid))
            video_result = video_meta_resp.json()
            if video_result and len(video_result) > 0:
                vidible_metadata.update(video_meta_resp.json()[0])
    return vidible_metadata


@bp.route('/vidible/bcid/<bcid>/pid/<pid>', methods=['GET', 'OPTIONS'])
def vidible(**kwargs):
    meta = get_vidible_metadata(**kwargs)
    if meta:
        return send_response(None, (meta, utcnow(), None, 200))
    else:
        return send_response(None, ({'error': 'not found'}, utcnow(), None, 404))



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import unittest
from .length_feet_and_inches_to_metric import feet_inches_to_metric


class FeetInchesTestCase(unittest.TestCase):

    def test_feet_to_cm(self):
        text = '''His height is 5'10" today
                His height is 5'10 today
                His height is 5' 10 today
                His height is 5 ft 10 today
                His height is 5 ft. 10 today
                His height is 5 ft 10 in today
                His height is 5' 10" today
                His height is 5' today
                His height is 5 ft today
                His height is 5 feet today
                His height is 10.2-ft today
                His height is 10.2-in today
                His height is 1,020" today
                His height is 1,020 in today
                His height is 1 inch today
                His height is 1,020 inches today
                His height is 10-12 ft today
                His height is 10-12 in today
                His height is 5-foot-10 today
                His height is 5-feet-10 today'''

        item = {'body_html': text}
        res, diff = feet_inches_to_metric(item)
        self.assertEqual(diff['5\'10"'], '5\'10" (1.78 m)')
        self.assertEqual(diff['5\'10'], '5\'10 (1.78 m)')
        self.assertEqual(diff['5\' 10'], '5\' 10 (1.78 m)')
        self.assertEqual(diff['5 ft 10'], '5 ft 10 (1.78 m)')
        self.assertEqual(diff['5 ft. 10'], '5 ft. 10 (1.78 m)')
        self.assertEqual(diff['5 ft 10 in'], '5 ft 10 in (1.78 m)')
        self.assertEqual(diff['5\' 10"'], '5\' 10" (1.78 m)')
        self.assertEqual(diff['5\''], '5\' (1.52 m)')
        self.assertEqual(diff['5 ft'], '5 ft (1.52 m)')
        self.assertEqual(diff['5 feet'], '5 feet (1.52 m)')
        self.assertEqual(diff['10.2-ft'], '10.2-ft (3.11 m)')
        self.assertEqual(diff['10.2-in'], '10.2-in (25.91 cm)')
        self.assertEqual(diff['1,020"'], '1,020" (25.91 m)')
        self.assertEqual(diff['1,020 in'], '1,020 in (25.91 m)')
        self.assertEqual(diff['1 inch'], '1 inch (2.54 cm)')
        self.assertEqual(diff['1,020 inches'], '1,020 inches (25.91 m)')
        self.assertEqual(diff['10-12 ft'], '10-12 ft (3.05-3.66 m)')
        self.assertEqual(diff['10-12 in'], '10-12 in (25.40-30.48 cm)')
        self.assertEqual(diff['5-foot-10'], '5-foot-10 (1.78 m)')
        self.assertEqual(diff['5-feet-10'], '5-feet-10 (1.78 m)')
        self.assertEqual(res['body_html'], item['body_html'])



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import re
import superdesk

from bson import ObjectId
from bson.errors import InvalidId
from cerberus import errors
from eve.io.mongo import Validator
from eve.utils import config
from eve.auth import auth_field_and_value
from flask import current_app as app
from flask_babel import _
from eve.validation import SingleErrorAsStringErrorHandler
from werkzeug.datastructures import FileStorage


ERROR_PATTERN = 'pattern'
ERROR_UNIQUE = 'unique'
ERROR_MINLENGTH = 'minlength'
ERROR_REQUIRED = 'required'
ERROR_JSON_LIST = 'json_list'

CLIENT_ERRORS = (
    ERROR_UNIQUE,
    ERROR_PATTERN,
    ERROR_REQUIRED,
    ERROR_MINLENGTH,
    ERROR_JSON_LIST,
)


class BaseErrorHandler(SingleErrorAsStringErrorHandler):
    def _unpack_single_element_lists(self, tree):
        for field in tree:
            error_list = tree[field]
            if len(error_list) > 0 and isinstance(tree[field][-1], dict):
                self._unpack_single_element_lists(tree[field][-1])
                # if there are sub field errors only return these for now
                if len(error_list) and any([isinstance(err, dict) for err in error_list]):
                    _errors = {}
                    for err in error_list:
                        if isinstance(err, dict):
                            _errors.update(err)
                    tree[field] = _errors
            if len(tree[field]) == 1 and isinstance(tree[field], list):
                tree[field] = tree[field][0]


class SuperdeskErrorHandler(BaseErrorHandler):
    def _format_message(self, field, error):
        if error.info and error.info[0] in CLIENT_ERRORS:
            return {error.info[0]: [1]}  # value must be list, will be unpacked
        elif error.code == errors.REQUIRED_FIELD.code:
            return {ERROR_REQUIRED: [1]}
        return self.messages[error.code].format(
            *error.info, constraint=error.constraint, field=field, value=error.value
        )


class SuperdeskValidator(Validator):

    def __init__(self, *args, **kwargs):
        kwargs['error_handler'] = SuperdeskErrorHandler
        super(SuperdeskValidator, self).__init__(*args, **kwargs)

    def _validate_mapping(self, mapping, field, value):
        """
        {'type': 'dict', 'nullable': True}
        """
        pass

    def _validate_index(self, definition, field, value):
        """
        {'type': 'string'}
        """
        pass

    def _validate_type_phone_number(self, value):
        """Enables validation for `phone_number` schema attribute.

        :param field: field name.
        :param value: field value.
        """
        if re.match("^(?:(?:0?[1-9][0-9]{8})|(?:(?:\+|00)[1-9][0-9]{9,11}))$", value):
            return True

    def _validate_type_email(self, value):
        """Enables validation for `email` schema attribute.

        :param field: field name.
        :param value: field value.
        """
        # it's tricky to write proper regex for email validation, so we
        # should use simple one, or use libraries like
        # - https://pypi.python.org/pypi/email_validator
        # - https://pypi.python.org/pypi/pyIsEmail
        # given that admins are usually create users, not users by themself,
        # probably just check for @ is enough
        # https://davidcel.is/posts/stop-validating-email-addresses-with-regex/
        if re.match('.+@.+', value, re.IGNORECASE):
            return True

    def _validate_type_file(self, value):
        """Enables validation for `file` schema attribute."""
        if isinstance(value, FileStorage):
            return True

    def _validate_multiple_emails(self, multiple, field, value):
        """
        {'type': 'boolean'}
        """
        if multiple:
            emails = value.split(',')
            if not all([self._validate_type_email(email) for email in emails]):
                self._error(field, ERROR_PATTERN)

    def _validate_unique(self, unique, field, value):
        """
        {'type': 'boolean'}
        """
        if not self.resource.endswith("autosave") and unique:
            query = {field: value}
            self._set_id_query(query)
            conflict = superdesk.get_resource_service(self.resource).find_one(req=None, **query)
            if conflict:
                self._error(field, ERROR_UNIQUE)

    def _set_id_query(self, query):
        if self.document_id:
            try:
                query[config.ID_FIELD] = {'$ne': ObjectId(self.document_id)}
            except InvalidId:
                query[config.ID_FIELD] = {'$ne': self.document_id}

    def _validate_iunique(self, unique, field, value):
        """
        {'type': 'boolean'}
        """
        if unique:
            pattern = '^{}$'.format(re.escape(value.strip()))
            query = {field: re.compile(pattern, re.IGNORECASE)}
            self._set_id_query(query)
            cursor = superdesk.get_resource_service(self.resource).get_from_mongo(req=None, lookup=query)
            if cursor.count():
                self._error(field, ERROR_UNIQUE)

    def _validate_iunique_per_parent(self, parent_field, field, value):
        """
        {'type': 'string'}
        """
        original = self.persisted_document or {}
        update = self.document or {}

        parent_field_value = update.get(parent_field, original.get(parent_field))

        if parent_field:
            pattern = '^{}$'.format(re.escape(value.strip()))
            query = {
                field: re.compile(pattern, re.IGNORECASE),
                parent_field: parent_field_value
            }
            self._set_id_query(query)

            cursor = superdesk.get_resource_service(self.resource).get_from_mongo(req=None, lookup=query)
            if cursor.count():
                self._error(field, ERROR_UNIQUE)

    def _validate_minlength(self, min_length, field, value):
        """
        {'type': 'integer'}
        """
        if isinstance(value, (type(''), list)):
            if len(value) < min_length:
                self._error(field, ERROR_MINLENGTH)

    def _validate_required_fields(self, document):
        """
        {'type': 'list'}
        """
        required = list(field for field, definition in self.schema.items()
                        if definition.get('required') is True)
        missing = set(required) - set(key for key in document.keys()
                                      if document.get(key) is not None or
                                      not self.ignore_none_values)
        for field in missing:
            self._error(field, ERROR_REQUIRED)

    def _validate_type_json_list(self, field, value):
        """It will fail later when loading."""
        if not isinstance(value, type('')):
            self._error(field, ERROR_JSON_LIST)

    def _validate_unique_to_user(self, unique, field, value):
        """
        {'type': 'boolean'}
        """
        doc = getattr(self, 'document', getattr(self, 'original_document', {}))

        if 'user' in doc:
            _, auth_value = auth_field_and_value(self.resource)
            query = {'user': auth_value}
        else:
            query = {'user': {'$exists': False}}

        self._is_value_unique(unique, field, value, query)

    def _validate_unique_template(self, unique, field, value):
        """
        {'type': 'boolean'}
        """
        original = self.persisted_document or {}
        update = self.document or {}

        is_public = update.get('is_public', original.get('is_public', None))
        template_name = update.get('template_name', original.get('template_name', None))

        if is_public:
            query = {'is_public': True}
        else:
            _, auth_value = auth_field_and_value(self.resource)
            query = {'user': auth_value, 'is_public': False}

        query['template_name'] = re.compile('^{}$'.format(re.escape(template_name.strip())), re.IGNORECASE)

        if self.document_id:
            id_field = config.DOMAIN[self.resource]['id_field']
            query[id_field] = {'$ne': self.document_id}

        if superdesk.get_resource_service(self.resource).find_one(req=None, **query):
            self._error(field, "Template Name is not unique")

    def _validate_twitter(self, twitter, field, value):
        """
        {'type': 'boolean'}
        """
        if twitter and value and not re.match('^@[A-Za-z0-9_]{1,15}$', value, re.IGNORECASE):
            self._error(field, ERROR_PATTERN)

    def _validate_username_pattern(self, enabled, field, value):
        """
        {'type': 'boolean'}
        """
        if enabled and app.config.get('USER_USERNAME_PATTERN') and \
                not re.match(app.config['USER_USERNAME_PATTERN'], value or ''):
            self._error(field, ERROR_PATTERN)

    def _validate_empty(self, empty, field, value):
        """
        {'type': 'boolean'}
        """
        # let the standard validation happen
        super()._validate_empty(empty, field, value)

        # custom validation
        if isinstance(value, list) or isinstance(value, dict):
            if len(value) == 0 and not empty:
                self._error(field, errors.EMPTY_NOT_ALLOWED)

    def _validate_unique_list(self, unique_list, field, value):
        """
        {'type': 'boolean'}
        """

        if unique_list and isinstance(value, list):
            if len(set(value)) != len(value):
                self._error(field, "Must contain unique items only.")

    def _validate_content_type_single_item_type(self, checked, field, value):
        """
        {'type': 'boolean'}
        """
        if checked and value not in {'text', None}:
            if app.data.find_one('content_types', req=None, item_type=value) is not None:
                self._error(field, _("Only 1 instance is allowed."))



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import json
from unittest import mock
from datetime import timedelta

from superdesk.tests import TestCase
from superdesk.publish.formatters.ninjs_ftp_formatter import FTPNinjsFormatter
from superdesk.publish import init_app
from superdesk.utc import utcnow


@mock.patch('superdesk.publish.subscribers.SubscribersService.generate_sequence_number', lambda self, subscriber: 1)
class FTPNinjsFormatterTest(TestCase):
    def setUp(self):
        self.formatter = FTPNinjsFormatter()
        init_app(self.app)
        self.maxDiff = None

    def test_picture_formatter(self):
        self.app.data.insert(
            "vocabularies",
            [
                {
                    "_id": "crop_sizes",
                    "items": [
                        {"is_active": True, "name": "custom", "width": 10000, "height": 10000},
                    ],
                }
            ],
        )

        embargoed = utcnow() + timedelta(days=2)
        article = {
            "guid": "20150723001158606583",
            "_current_version": 1,
            "slugline": "AMAZING PICTURE",
            "original_source": "AAP",
            "renditions": {
                "viewImage": {
                    "width": 640,
                    "href": "http://localhost:5000/api/upload/55b032041d41c8d278d21b6f/raw?_schema=http",
                    "mimetype": "image/jpeg",
                    "height": 401,
                },
                "original": {
                    "href": "https://one-api.aap.com.au/api/v3/Assets/20150723001158606583/Original/download",
                    "mimetype": "image/jpeg",
                    "media": "12345/5e448dd1016d1f63a92f0398.jpg"
                },
                "custom": {
                    "href": "https://one-api.aap.com.au/api/v3/Assets/20150723001158637458/Original/download",
                    "mimetype": "image/jpeg",
                    "media": "12345/5e448dd1016d1f63a92f0400.jpg"
                }
            },
            "byline": "MICKEY MOUSE",
            "headline": "AMAZING PICTURE",
            "versioncreated": "2015-07-23T00:15:00.000Z",
            "ednote": "TEST ONLY",
            "type": "picture",
            "pubstatus": "usable",
            "source": "AAP",
            "description": "The most amazing picture you will ever see",
            "guid": "20150723001158606583",
            "body_footer": "<p>call helpline 999 if you are planning to quit smoking</p>",
            "embargoed": embargoed,
        }
        seq, doc = self.formatter.format(article, {"name": "Test Subscriber", "destinations":
                                                   [{"config": {"host": "ftp.abc.com", "path": "/stories",
                                                                "associated_path": "/pictures",
                                                                "push_associated": True}}]})[0]
        expected = {
            "byline": "MICKEY MOUSE",
            "renditions": {
                "custom": {
                    "href": "/pictures/12345-5e448dd1016d1f63a92f0400.jpg",
                    "mimetype": "image/jpeg",
                    "media": "12345/5e448dd1016d1f63a92f0400.jpg"
                },
            },
            "headline": "AMAZING PICTURE",
            "pubstatus": "usable",
            "version": "1",
            "versioncreated": "2015-07-23T00:15:00.000Z",
            "guid": "20150723001158606583",
            "description_html": "The most amazing picture you will ever see<p>call helpline 999 if you are planning to "
                                "quit smoking</p>",
            "type": "picture",
            "priority": 5,
            "slugline": "AMAZING PICTURE",
            "ednote": "TEST ONLY",
            "source": "AAP",
            "embargoed": embargoed.isoformat(),
        }
        self.assertEqual(expected, json.loads(doc))
        self.assertNotIn('viewImage', json.loads(doc).get('renditions'))

    def test_picture_formatter_with_original(self):
        self.app.data.insert(
            "vocabularies",
            [
                {
                    "_id": "crop_sizes",
                    "items": [
                        {"is_active": True, "name": "custom", "width": 10000, "height": 10000},
                    ],
                }
            ],
        )

        embargoed = utcnow() + timedelta(days=2)
        article = {
            "guid": "20150723001158606583",
            "_current_version": 1,
            "slugline": "AMAZING PICTURE",
            "original_source": "AAP",
            "renditions": {
                "viewImage": {
                    "width": 640,
                    "href": "http://localhost:5000/api/upload/55b032041d41c8d278d21b6f/raw?_schema=http",
                    "mimetype": "image/jpeg",
                    "height": 401,
                },
                "original": {
                    "href": "https://one-api.aap.com.au/api/v3/Assets/20150723001158606583/Original/download",
                    "mimetype": "image/jpeg",
                    "media": "12345/5e448dd1016d1f63a92f0398.jpg"
                },
                "custom": {
                    "href": "https://one-api.aap.com.au/api/v3/Assets/20150723001158637458/Original/download",
                    "mimetype": "image/jpeg",
                    "media": "12345/5e448dd1016d1f63a92f0400.jpg"
                }
            },
            "byline": "MICKEY MOUSE",
            "headline": "AMAZING PICTURE",
            "versioncreated": "2015-07-23T00:15:00.000Z",
            "ednote": "TEST ONLY",
            "type": "picture",
            "pubstatus": "usable",
            "source": "AAP",
            "description": "The most amazing picture you will ever see",
            "guid": "20150723001158606583",
            "body_footer": "<p>call helpline 999 if you are planning to quit smoking</p>",
            "embargoed": embargoed,
        }
        seq, doc = self.formatter.format(article, {"name": "Test Subscriber", "destinations":
                                                   [{"config": {"host": "ftp.abc.com", "path": "/stories",
                                                                "associated_path": "/pictures",
                                                                "push_associated": True,
                                                                "include_original": True}}]})[0]
        expected = {
            "byline": "MICKEY MOUSE",
            "renditions": {
                "custom": {
                    "href": "/pictures/12345-5e448dd1016d1f63a92f0400.jpg",
                    "mimetype": "image/jpeg",
                    "media": "12345/5e448dd1016d1f63a92f0400.jpg"
                },
                "original": {
                    "href": "/pictures/12345-5e448dd1016d1f63a92f0398.jpg",
                    "mimetype": "image/jpeg",
                    "media": "12345/5e448dd1016d1f63a92f0398.jpg"
                },
            },
            "headline": "AMAZING PICTURE",
            "pubstatus": "usable",
            "version": "1",
            "versioncreated": "2015-07-23T00:15:00.000Z",
            "guid": "20150723001158606583",
            "description_html": "The most amazing picture you will ever see<p>call helpline 999 if you are planning to "
                                "quit smoking</p>",
            "type": "picture",
            "priority": 5,
            "slugline": "AMAZING PICTURE",
            "ednote": "TEST ONLY",
            "source": "AAP",
            "embargoed": embargoed.isoformat(),
        }
        self.assertEqual(expected, json.loads(doc))
        self.assertNotIn('viewImage', json.loads(doc).get('renditions'))

    def test_embedded_image(self):
        self.app.data.insert(
            "vocabularies",
            [
                {
                    "_id": "crop_sizes",
                    "items": [
                        {"is_active": True, "name": "custom", "width": 10000, "height": 10000},
                    ],
                }
            ],
        )

        embargoed = utcnow() + timedelta(days=2)
        article = {
            "_id": "urn:newsml:localhost:2020-03-12T15:19:39.654956:e78f3dd6-c096-43d5-9ba0-014e07dc4f1f",
            "target_regions": [],
            "priority": 6,
            "place": [
                {
                    "world_region": "Asia",
                    "name": "ASIA",
                    "country": "",
                    "qcode": "ASIA",
                    "group": "Rest Of World",
                    "state": ""
                }
            ],
            "associations": {
                "editor_0": {
                    "_current_version": 1,
                    "byline": "Some One",
                    "_latest_version": 1,
                    "family_id": "20181226001377597230",
                    "type": "picture",
                    "unique_id": 35167342,
                    "_id": "tag:localhost:2019:5c9da5e1-f491-4ffd-a0b5-dd54e9fd870f",
                    "versioncreated": "2019-04-05T05:37:44+0000",
                    "alt_text": "Power plant",
                    "archive_description": "Steam rises from the brown coal-fired power plant",
                    "format": "HTML",
                    "pubstatus": "usable",
                    "mimetype": "image/jpeg",
                    "description_text": "Steam rises from the brown coal-fired power plant",
                    "firstcreated": "2018-12-04T15:17:37+0000",
                    "renditions": {
                        "original": {
                            "mimetype": "image/jpeg",
                            "href": "http://superdesk-uat.aap.com.au/api/upload-raw/20190405160436/9e0b3096a5792b1e6"
                                    "46a6b1e7dad606ccc697b4b686dde76b13b7b64e994f579.jpg",
                            "width": 4500,
                            "height": 3000,
                            "poi": {
                                "y": 1770,
                                "x": 2160
                            },
                            "media": "20190405160436/9e0b3096a5792b1e646a6b1e7dad606ccc"
                                     "697b4b686dde76b13b7b64e994f579.jpg"
                        }
                    },
                    "language": "en",
                    "state": "fetched",
                    "_links": {
                        "collection": {
                            "title": "archive",
                            "href": "archive"
                        },
                        "parent": {
                            "title": "home",
                            "href": "/"
                        },
                        "self": {
                            "title": "Archive",
                            "href": "archive/tag:localhost:2019:5c9da5e1-f491-4ffd-a0b5-dd54e9fd870f"
                        }
                    },
                    "_etag": "16d959004ace3137f977650b52ba1db74ca4058b",
                    "source": "XXX",
                    "ingest_id": "20181226001377597230",
                    "original_source": "XXX/XXX",
                    "priority": 6,
                    "operation": "fetch",
                    "sign_off": "MD",
                    "version_creator": "5673a749069b7f6e7ff8de6c",
                    "guid": "20181226001377597230",
                    "expiry": "2019-04-08T05:37:45+0000",
                    "headline": "YEARENDER DECEMBER 2018",
                    "original_creator": "5673a749069b7f6e7ff8de6c",
                    "urgency": 3,
                    "slugline": "YEARENDER DECEMBER 2018",
                    "genre": [
                        {
                            "qcode": "Article",
                            "name": "Article"
                        }
                    ],
                    "poi": {
                        "y": 0.59,
                        "x": 0.48
                    }
                }
            },
            "family_id": "urn:newsml:localhost:2020-03-12T15:19:39.654956:e78f3dd6-c096-43d5-9ba0-014e07dc4f1f",
            "keywords": [],
            "genre": [
                {
                    "qcode": "Article",
                    "name": "Article",
                    "scheme": None
                }
            ],
            "format": "HTML",
            "unique_id": 37642265,
            "operation": "update",
            "type": "text",
            "description_text": '',
            "headline": "Test Headline",
            "urgency": 3,
            "unique_name": "#37642265",
            "anpa_category": [
                {
                    "qcode": "a",
                    "name": "Australian General News"
                }
            ],
            "version": 1,
            "byline": "Fred Flinstone",
            "target_subscribers": [],
            "slugline": "Test",
            "subject": [
                {
                    "qcode": "01001000",
                    "name": "archaeology",
                    "parent": "01000000"
                }
            ],
            "language": "en",
            "body_html": "<p>Some text</p>\n<!-- EMBED START Image {id: \"editor_0\"} -->\n<figure>\n    "
                         "<img src=\"http://superdesk-uat.aap.com.au/api/upload-raw/20190405160436/9e0b3096a5792b1"
                         "e646a6b1e7dad606ccc697b4b686dde76b13b7b64e994f579.jpg\" alt=\"Power plant\" />\n    "
                         "<figcaption>Steam rises from the brown coal-fired power plant</figcaption>\n</figure>\n"
                         "<!-- EMBED END Image {id: \"editor_0\"} -->\n<p>More text</p>",
            "source": "aap",
            "body_footer": "",
            "guid": "urn:newsml:localhost:2020-03-12T15:19:39.654956:e78f3dd6-c096-43d5-9ba0-014e07dc4f1f",

        }
        seq, doc = self.formatter.format(article, {"name": "Test Subscriber", "destinations":
                                                   [{"config": {"host": "ftp.abc.com", "path": "/stories",
                                                                "associated_path": "/pictures",
                                                                "push_associated": True,
                                                                "include_original": True}}]})[0]
        expected = {
            'source': 'aap', 'readtime': 0, 'headline': 'Test Headline', 'keywords': [], 'associations': {
                'editor_0': {'source': 'XXX', 'headline': 'YEARENDER DECEMBER 2018',
                             'firstcreated': '2018-12-04T15:17:37+0000', 'guid': '20181226001377597230',
                             'type': 'picture', 'byline': 'Some One', 'urgency': 3, 'language': 'en',
                             'slugline': 'YEARENDER DECEMBER 2018', 'versioncreated': '2019-04-05T05:37:44+0000',
                             'mimetype': 'image/jpeg', 'genre': [{'name': 'Article', 'code': 'Article'}], 'priority': 6,
                             'renditions': {
                                 'original': {'mimetype': 'image/jpeg', 'poi': {'x': 2160, 'y': 1770}, 'height': 3000,
                                              'media': '20190405160436/9e0b3096a5792b1e646a6b1e7dad606ccc697b4b686'
                                                       'dde76b13b7b64e994f579.jpg',
                                              'width': 4500,
                                              'href': '/pictures/20190405160436-9e0b3096a5792b1e646a6b1e7dad606ccc69'
                                                      '7b4b686dde76b13b7b64e994f579.jpg'}},
                             'description_text': 'Steam rises from the brown coal-fired power plant',
                             'pubstatus': 'usable', 'version': '1', 'body_text': 'Power plant'}},
            'subject': [{'name': 'archaeology', 'code': '01001000'}],
            'guid': 'urn:newsml:localhost:2020-03-12T15:19:39.654956:e78f3dd6-c096-43d5-9ba0-014e07dc4f1f',
            'slugline': 'Test', 'charcount': 67, 'type': 'text', 'urgency': 3, 'language': 'en',
            'byline': 'Fred Flinstone', 'wordcount': 12, 'genre': [{'name': 'Article', 'code': 'Article'}],
            'priority': 6, 'place': [{'name': 'ASIA', 'code': 'ASIA'}],
            'body_html': '<p>Some text</p>\n<!-- EMBED START Image {id: "editor_0"} -->\n<figure>\n   '
                         ' <img src="/pictures/20190405160436-9e0b3096a5792b1e646a6b1e7dad606ccc697b4b686dde76b13'
                         'b7b64e994f579.jpg" alt="Power plant" id="editor_0">\n    <figcaption>Steam rises from the'
                         ' brown coal-fired power plant</figcaption>\n</figure>\n<!-- EMBED END Image {id: "editor_0"}'
                         ' -->\n<p>More text</p>',
            'service': [{'name': 'Australian General News', 'code': 'a'}], 'version': '1'}
        self.assertEqual(expected, json.loads(doc))



from lxml import etree


from .base import BasePackageElement


class Preferences(BasePackageElement):
    """
    As Adobe's idml specification explains:
       ``
       The Preferences.xml file contains representations of all of the document preferences.
       ``
    """

    DOCUMENTPREFERENCE_DEFAULTS = {
        'PageHeight': '792',
        'PageWidth': '612',
        'PagesPerDocument': '1',
        'FacingPages': 'true',
        'DocumentBleedTopOffset': '0',
        'DocumentBleedBottomOffset': '0',
        'DocumentBleedInsideOrLeftOffset': '0',
        'DocumentBleedOutsideOrRightOffset': '0',
        'DocumentBleedUniformSize': 'true',
        'SlugTopOffset': '0',
        'SlugBottomOffset': '0',
        'SlugInsideOrLeftOffset': '0',
        'SlugRightOrOutsideOffset': '0',
        'DocumentSlugUniformSize': 'false',
        'PreserveLayoutWhenShuffling': 'true',
        'AllowPageShuffle': 'true',
        'OverprintBlack': 'true',
        'PageBinding': 'LeftToRight',
        'ColumnDirection': 'Horizontal',
        'Intent': 'PrintIntent'
    }

    @property
    def filename(self):
        """
        Filename inside IDML package.
        Used as a filename for a file inside zip container.
        :return str: filename
        """
        return 'Resources/Preferences.xml'

    def _build_etree(self):
        self._etree = etree.Element(
            etree.QName(self.XMLNS_IDPKG, 'Preferences'),
            nsmap={'idPkg': self.XMLNS_IDPKG}
        )
        self._etree.set('DOMVersion', self.DOM_VERSION)
        self._add_documentpreference()

    def _add_documentpreference(self):
        """
        Add <DocumentPreference..> and fill it with properties.
        :return lxml.etree: DocumentPreference
        """
        # DocumentPreference
        documentpreference = etree.SubElement(
            self._etree,
            'DocumentPreference',
            attrib=self.merge_attributes(
                self.DOCUMENTPREFERENCE_DEFAULTS,
                self._attributes.get('DocumentPreference', {})
            )
        )
        # Properties
        properties = etree.SubElement(
            documentpreference,
            'Properties'
        )
        # ColumnGuideColor
        columnguidecolor = etree.SubElement(
            properties,
            'ColumnGuideColor'
        )
        columnguidecolor.set('type', 'enumeration')
        columnguidecolor.text = 'Violet'
        # MarginGuideColor
        marginguidecolor = etree.SubElement(
            properties,
            'MarginGuideColor'
        )
        marginguidecolor.set('type', 'enumeration')
        marginguidecolor.text = 'Magenta'

        return documentpreference




import socket
import ftplib

from contextlib import contextmanager
from flask import current_app as app

from superdesk.errors import IngestFtpError


@contextmanager
def ftp_connect(config):
    """Get ftp connection for given config.

    use with `with`

    :param config: dict with `host`, `username`, `password`, `path` and `passive`
    """
    try:
        ftp = ftplib.FTP(config.get('host'), timeout=app.config.get('FTP_TIMEOUT', 300))
    except socket.gaierror as e:
        raise IngestFtpError.ftpHostError(exception=e)
    if config.get('username'):
        try:
            ftp.login(config.get('username'), config.get('password'))
        except ftplib.error_perm as e:
            raise IngestFtpError.ftpAuthError(exception=e)
    if config.get('path'):
        ftp.cwd(config.get('path', '').lstrip('/'))
    if config.get('passive') is False:  # only set this when not active, it's passive by default
        ftp.set_pasv(False)
    yield ftp
    ftp.close()



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import unittest
from .area_square_yards_to_metric import square_yard_to_metric


class TemperatureTestCase(unittest.TestCase):

    def test_f_to_c(self):
        text = 'Total area is 100.50 square yards for this land. '\
               'Total area is 15.7 square yds for this land. '\
               'Total area is 1 sq.yd for this land. '\
               'Total area is 1-16 sq-yds for this land. '\
               'Total area is 16.7-Square-yds for this land. '\
               'Total area is 16,500-Square Yards. for this land. '\

        item = {'body_html': text}
        res, diff = square_yard_to_metric(item)
        self.assertEqual(diff['100.50 square yards'], '100.50 square yards (84.03 square meter)')
        self.assertEqual(diff['15.7 square yds'], '15.7 square yds (13.1 square meter)')
        self.assertEqual(diff['1 sq.yd'], '1 sq.yd (0.84 square meter)')
        self.assertEqual(diff['1-16 sq-yds'], '1-16 sq-yds (0.84-13 square meter)')
        self.assertEqual(diff['16.7-Square-yds'], '16.7-Square-yds (14.0 square meter)')
        self.assertEqual(diff['16,500-Square Yards'], '16,500-Square Yards (1.4 ha)')
        self.assertEqual(res['body_html'], item['body_html'])



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import unittest
from .length_miles_to_metric import miles_to_metric


class MilesTestCase(unittest.TestCase):

    def test_m_to_km(self):
        text = 'Tom rides 125 miles everyday. ' \
               'Tom rides 1 mile everyday. ' \
               'Tom rides 125 Miles everyday. ' \
               'Tom rides 125-Miles everyday. ' \
               'Tom rides 12.5 mi everyday. ' \
               'Tom rides 12.5mi everyday. ' \
               'Tom rides 12.5-mi everyday. ' \
               'Tom rides 12,500 miles everyday. ' \
               'Tom rides 100-12500 Miles everyday. ' \

        item = {'body_html': text}
        res, diff = miles_to_metric(item)
        self.assertEqual(diff['125 miles'], '125 miles (201.2 km)')
        self.assertEqual(diff['1 mile'], '1 mile (1.6 km)')
        self.assertEqual(diff['125 Miles'], '125 Miles (201.2 km)')
        self.assertEqual(diff['125-Miles'], '125-Miles (201.2 km)')
        self.assertEqual(diff['12.5 mi'], '12.5 mi (20.1 km)')
        self.assertEqual(diff['12.5mi'], '12.5mi (20.1 km)')
        self.assertEqual(diff['12.5-mi'], '12.5-mi (20.1 km)')
        self.assertEqual(diff['12,500 miles'], '12,500 miles (20,117 km)')
        self.assertEqual(diff['100-12500 Miles'], '100-12500 Miles (160.9-20,117 km)')
        self.assertEqual(res['body_html'], item['body_html'])



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import io
import json
import logging

from bson import ObjectId
from functools import partial
import content_api
from flask import current_app as app
from superdesk import get_resource_service
from superdesk.errors import SuperdeskApiError, SuperdeskPublishError
from superdesk.metadata.item import CONTENT_TYPE, ITEM_TYPE, ITEM_STATE, PUBLISH_SCHEDULE, ASSOCIATIONS
from superdesk.metadata.packages import GROUPS, ROOT_GROUP, GROUP_ID, REFS, RESIDREF
from superdesk.notification import push_notification
from superdesk.publish import SUBSCRIBER_TYPES
from superdesk.publish.publish_queue import PUBLISHED_IN_PACKAGE
from superdesk.publish.formatters import get_formatter
from apps.publish.content.common import BasePublishService
from copy import deepcopy
from eve.utils import config, ParsedRequest
from apps.archive.common import get_user, get_utc_schedule
from apps.packages.package_service import PackageService
from apps.publish.published_item import PUBLISH_STATE, QUEUE_STATE
from apps.content_types import apply_schema
from datetime import datetime
import pytz
from flask_babel import _

logger = logging.getLogger(__name__)


class EnqueueService:
    """
    Creates the corresponding entries in the publish queue for items marked for publishing
    """

    publish_type = 'publish'
    published_state = 'published'

    non_digital = partial(filter, lambda s: s.get('subscriber_type', '') == SUBSCRIBER_TYPES.WIRE)
    digital = partial(filter, lambda s: (s.get('subscriber_type', '') in {SUBSCRIBER_TYPES.DIGITAL,
                                                                          SUBSCRIBER_TYPES.ALL}))
    package_service = PackageService()

    filters = None

    def __init__(self, published_state=None):
        if published_state is not None:
            self.published_state = published_state

    def get_filters(self):
        """Retrieve all of the available filter conditions and content filters if they have not yet been retrieved or
        they have been updated. This avoids the filtering functions having to repeatedly retireve the individual filter
        records.

        :return:
        """

        # Get the most recent update time to the filter conditions and content_filters
        req = ParsedRequest()
        req.sort = '-_updated'
        req.max_results = 1
        mindate = datetime.min.replace(tzinfo=pytz.UTC)
        latest_fc = next(get_resource_service('filter_conditions').get_from_mongo(req=req, lookup=None),
                         {}).get('_updated', mindate)
        latest_cf = next(get_resource_service('content_filters').get_from_mongo(req=req, lookup=None),
                         {}).get('_updated', mindate)

        if not self.filters or \
                latest_fc > self.filters.get('latest_filter_conditions', mindate) or latest_fc == mindate or \
                latest_cf > self.filters.get('latest_content_filters', mindate) or latest_cf == mindate:
            logger.debug('Getting content filters and filter conditions')
            self.filters = dict()
            self.filters['filter_conditions'] = dict()
            self.filters['content_filters'] = dict()
            for fc in get_resource_service('filter_conditions').get(req=None, lookup={}):
                self.filters['filter_conditions'][fc.get('_id')] = {'fc': fc}
                self.filters['latest_filter_conditions'] = fc.get('_updated') if fc.get('_updated') > self.filters.get(
                    'latest_filter_conditions', mindate) else self.filters.get('latest_filter_conditions', mindate)
            for cf in get_resource_service('content_filters').get(req=None, lookup={}):
                self.filters['content_filters'][cf.get('_id')] = {'cf': cf}
                self.filters['latest_content_filters'] = cf.get('_updated') if cf.get('_updated') > self.filters.get(
                    'latest_content_filters', mindate) else self.filters.get('latest_content_filters', mindate)
        else:
            logger.debug('Using chached content filters and filters conditions')

    def _enqueue_item(self, item, content_type=None):
        item_to_queue = deepcopy(item)
        if item[ITEM_TYPE] == CONTENT_TYPE.COMPOSITE:
            queued = self._publish_package_items(item_to_queue)
            if not queued:  # this was only published to subscribers with config.packaged on
                return self.publish(doc=item_to_queue, target_media_type=SUBSCRIBER_TYPES.DIGITAL)
            else:
                return queued
        elif content_type:
            return self.publish(item_to_queue, None, content_type)
        elif item[ITEM_TYPE] not in [CONTENT_TYPE.TEXT, CONTENT_TYPE.PREFORMATTED]:
            return self.publish(item_to_queue, SUBSCRIBER_TYPES.DIGITAL)
        else:
            return self.publish(item_to_queue, None)

    def _publish_package_items(self, package):
        """Publishes all items of a package recursively then publishes the package itself

        :param package: package to publish
        :param updates: payload
        """
        items = self.package_service.get_residrefs(package)
        subscriber_items = {}
        queued = False
        removed_items = []
        if self.publish_type in ['correct', 'kill']:
            removed_items, added_items = self._get_changed_items(items, package)
            # we raise error if correction is done on a empty package. Kill is fine.
            if len(removed_items) == len(items) and len(added_items) == 0 and self.publish_type == 'correct':
                raise SuperdeskApiError.badRequestError(_("Corrected package cannot be empty!"))
            items.extend(added_items)

        if items:
            archive_service = get_resource_service('archive')
            for guid in items:
                package_item = archive_service.find_one(req=None, _id=guid)

                if not package_item:
                    raise SuperdeskApiError.badRequestError(
                        _("Package item with id: {guid} has not been published.").format(guid=guid))

                subscribers, subscriber_codes, associations = self._get_subscribers_for_package_item(package_item)
                package_item_id = package_item[config.ID_FIELD]
                self._extend_subscriber_items(subscriber_items,
                                              subscribers,
                                              package_item,
                                              package_item_id,
                                              subscriber_codes)

            for removed_id in removed_items:
                package_item = archive_service.find_one(req=None, _id=removed_id)
                subscribers, subscriber_codes, associations = self._get_subscribers_for_package_item(package_item)
                package_item_id = None
                self._extend_subscriber_items(subscriber_items,
                                              subscribers,
                                              package_item,
                                              package_item_id,
                                              subscriber_codes)

            queued = self.publish_package(package, target_subscribers=subscriber_items)

        return queued

    def _get_changed_items(self, existing_items, package):
        """Returns the added and removed items from existing_items

        :param existing_items: Existing list
        :param updates: Changes
        :return: list of removed items and list of added items
        """
        published_service = get_resource_service('published')
        req = ParsedRequest()
        query = {'query': {'filtered': {'filter': {'and': [{'terms': {QUEUE_STATE: [
            PUBLISH_STATE.QUEUED, PUBLISH_STATE.QUEUED_NOT_TRANSMITTED]}},
            {'term': {'item_id': package['item_id']}}]}}}, 'sort': [{'publish_sequence_no': 'desc'}]}
        req.args = {'source': json.dumps(query)}
        req.max_results = 1
        previously_published_packages = published_service.get(req=req, lookup=None)

        if not previously_published_packages.count():
            return [], []

        previously_published_package = previously_published_packages[0]

        if 'groups' in previously_published_package:
            old_items = self.package_service.get_residrefs(previously_published_package)
            added_items = list(set(existing_items) - set(old_items))
            removed_items = list(set(old_items) - set(existing_items))
            return removed_items, added_items
        else:
            return [], []

    def enqueue_item(self, item, content_type=None):
        """Creates the corresponding entries in the publish queue for the given item

        :param item: Item to enqueue
        :param content_type: item content type
        :return bool: True if item is queued else false.
        """
        try:
            return self._enqueue_item(item, content_type)
        except SuperdeskApiError as e:
            raise e
        except KeyError as e:
            raise SuperdeskApiError.badRequestError(
                message=_("Key is missing on article to be published: {exception}").format(exception=str(e)))
        except Exception as e:
            logger.exception("Something bad happened while publishing {}".format(id))
            raise SuperdeskApiError.internalError(
                message=_("Failed to publish the item: {exception}").format(exception=str(e)),
                exception=e
            )

    def get_subscribers(self, doc, target_media_type):
        """Get subscribers for doc based on target_media_type.

        Override this method in the ArchivePublishService, ArchiveCorrectService and ArchiveKillService

        :param doc: Document to publish/correct/kill
        :param target_media_type: Valid values are - Wire, Digital.
        :return: (list, list) List of filtered subscriber,
                List of subscribers that have not received item previously (empty list in this case).
        """
        raise NotImplementedError()

    def publish(self, doc, target_media_type=None, content_type=None):
        """Queue the content for publishing.

        1. Get the subscribers.
        2. Queue the content for subscribers
        3. Sends notification if no formatter has found for any of the formats configured in Subscriber.
        4. If not queued and not formatters then raise exception.
        5. Publish the content to content api.

        :param dict doc: document to publish
        :param str target_media_type: Valid values are - Wire, Digital.
        :param str content_type: doc content type, None for content
        :return bool: if content is queued then True else False
        :raises PublishQueueError.item_not_queued_error:
                If the nothing is queued.
        """
        # Step 1
        subscribers, subscriber_codes, associations = self.get_subscribers(doc, target_media_type)

        # Step 2
        no_formatters, queued = self.queue_transmission(deepcopy(doc), subscribers, subscriber_codes, associations)

        # Step 3
        self._push_formatter_notification(doc, no_formatters)

        # Step 4
        if not target_media_type and not queued:
            level = logging.INFO
            if app.config['PUBLISH_NOT_QUEUED_ERROR'] and not app.config.get('SUPERDESK_TESTING'):
                level = logging.ERROR
            logger.log(level, 'Nothing is saved to publish queue for story: {} for action: {}'.
                       format(doc[config.ID_FIELD], self.publish_type))

        # Step 5
        if not content_type:
            self.publish_content_api(doc, [s for s in subscribers if s.get('api_enabled')])

        return queued

    def publish_content_api(self, doc, subscribers):
        """
        Publish item to content api
        :param dict doc: content api item
        :param list subscribers: list of subscribers
        """
        try:
            if content_api.is_enabled():
                get_resource_service('content_api').publish(doc, subscribers)
        except Exception:
            logger.exception('Failed to queue item to API for item: {} for action {}'.
                             format(doc[config.ID_FIELD], self.publish_type))

    def _push_formatter_notification(self, doc, no_formatters=None):
        if no_formatters is None:
            no_formatters = []

        if len(no_formatters) > 0:
            user = get_user()
            push_notification('item:publish:wrong:format',
                              item=str(doc[config.ID_FIELD]), unique_name=doc.get('unique_name'),
                              desk=str(doc.get('task', {}).get('desk', '')),
                              user=str(user.get(config.ID_FIELD, '')),
                              formats=no_formatters)

    def _get_subscriber_codes(self, subscribers):
        subscriber_codes = {}
        all_products = list(get_resource_service('products').get(req=None, lookup=None))

        for subscriber in subscribers:
            codes = self._get_codes(subscriber)
            products = [p for p in all_products if p[config.ID_FIELD] in subscriber.get('products', [])]

            for product in products:
                codes.extend(self._get_codes(product))
                subscriber_codes[subscriber[config.ID_FIELD]] = list(set(codes))

        return subscriber_codes

    def resend(self, doc, subscribers):
        """Resend doc to subscribers

        :param dict doc: doc to resend
        :param list subscribers: list of subscribers
        :return:
        """
        subscriber_codes = self._get_subscriber_codes(subscribers)
        wire_subscribers = list(self.non_digital(subscribers))
        digital_subscribers = list(self.digital(subscribers))

        for subscriber in wire_subscribers:
            subscriber['api_enabled'] = len(subscriber.get('api_products') or []) > 0

        for subscriber in digital_subscribers:
            subscriber['api_enabled'] = len(subscriber.get('api_products') or []) > 0

        doc['item_id'] = doc[config.ID_FIELD]
        associations = self._resend_associations_to_subscribers(doc, subscribers)
        if len(wire_subscribers) > 0:
            self._resend_to_subscribers(doc, wire_subscribers, subscriber_codes, associations)
            self.publish_content_api(doc,
                                     [subscriber for subscriber in wire_subscribers
                                      if subscriber.get('api_enabled')])

        if len(digital_subscribers) > 0:
            package = None
            self._resend_to_subscribers(doc, digital_subscribers, subscriber_codes, associations)

            self.publish_content_api(package or doc,
                                     [subscriber for subscriber in digital_subscribers
                                      if subscriber.get('api_enabled')])

    def _resend_associations_to_subscribers(self, doc, subscribers):
        """
        On resend association are also sent to the subscribers.
        :param dict doc: item to resend
        :param list subscribers: list of subscribers
        :return dict: associations
        """
        if not doc.get(ASSOCIATIONS):
            return {}

        associations = {}

        for assoc_id, item in doc.get(ASSOCIATIONS).items():
            if not item:
                continue

            item['subscribers'] = []

            for s in subscribers:
                item['subscribers'].append(s.get(config.ID_FIELD))
                if not associations.get(s.get(config.ID_FIELD)):
                    associations[s.get(config.ID_FIELD)] = []

                associations[s.get(config.ID_FIELD)].append(item.get(config.ID_FIELD))
        return associations

    def _resend_to_subscribers(self, doc, subscribers, subscriber_codes, associations=None):
        if associations is None:
            associations = {}
        formatter_messages, queued = self.queue_transmission(doc, subscribers, subscriber_codes, associations)
        self._push_formatter_notification(doc, formatter_messages)
        if not queued:
            logger.exception('Nothing is saved to publish queue for story: {} for action: {}'.
                             format(doc[config.ID_FIELD], 'resend'))

    def publish_package(self, package, target_subscribers):
        """Publishes a given package to given subscribers.

        For each subscriber updates the package definition with the wanted_items for that subscriber
        and removes unwanted_items that doesn't supposed to go that subscriber.
        Text stories are replaced by the digital versions.

        :param package: Package to be published
        :param target_subscribers: List of subscriber and items-per-subscriber
        """
        all_items = self.package_service.get_residrefs(package)
        no_formatters, queued = [], False
        subscribers = []
        for items in target_subscribers.values():
            updated = deepcopy(package)
            subscriber = items['subscriber']
            codes = items['codes']
            wanted_items = [item for item in items['items'] if items['items'].get(item, None)]
            unwanted_items = [item for item in all_items if item not in wanted_items]
            for i in unwanted_items:
                still_items_left = self.package_service.remove_ref_from_inmem_package(updated, i)
                if not still_items_left and self.publish_type != 'correct':
                    # if nothing left in the package to be published and
                    # if not correcting then don't send the package
                    return
            for key in wanted_items:
                try:
                    self.package_service.replace_ref_in_package(updated, key, items['items'][key])
                except KeyError:
                    continue

            formatters, temp_queued = self.queue_transmission(updated, [subscriber],
                                                              {subscriber[config.ID_FIELD]: codes})

            subscribers.append(subscriber)
            no_formatters.extend(formatters)
            if temp_queued:
                queued = temp_queued

            delivery_types = [d['delivery_type'] for d in self.get_destinations(subscriber)]
            is_content_api_delivery = 'content_api' in delivery_types
            # packages for content_api will not be transmitted
            # so we need to publish them here
            if is_content_api_delivery and subscriber.get('api_enabled'):
                self.publish_content_api(package, [subscriber])

        return queued

    def get_destinations(self, subscriber):
        destinations = subscriber.get('destinations') or []
        if subscriber.get('api_enabled'):
            destinations.append({'name': 'content api', 'delivery_type': 'content_api', 'format': 'ninjs'})
        return destinations

    def queue_transmission(self, doc, subscribers, subscriber_codes=None, associations=None):
        """Method formats and then queues the article for transmission to the passed subscribers.

        ::Important Note:: Format Type across Subscribers can repeat. But we can't have formatted item generated once
        based on the format_types configured across for all the subscribers as the formatted item must have a published
        sequence number generated by Subscriber.

        :param dict doc: document to queue for transmission
        :param list subscribers: List of subscriber dict.
        :return : (list, bool) tuple of list of missing formatters and boolean flag. True if queued else False
        """

        if associations is None:
            associations = {}
        if subscriber_codes is None:
            subscriber_codes = {}

        try:
            queued = False
            no_formatters = []
            for subscriber in subscribers:
                try:
                    if doc[ITEM_TYPE] not in [CONTENT_TYPE.TEXT, CONTENT_TYPE.PREFORMATTED] and \
                            subscriber.get('subscriber_type', '') == SUBSCRIBER_TYPES.WIRE:
                        # wire subscribers can get only text and preformatted stories
                        continue

                    for destination in self.get_destinations(subscriber):
                        embed_package_items = doc[ITEM_TYPE] == CONTENT_TYPE.COMPOSITE and \
                            (destination.get('config') or {}).get('packaged', False)
                        if embed_package_items:
                            doc = self._embed_package_items(doc)

                        if doc.get(PUBLISHED_IN_PACKAGE) and \
                                (destination.get('config') or {}).get('packaged', False):
                            continue

                        # Step 2(a)
                        formatter = get_formatter(destination['format'], doc)

                        if not formatter:  # if formatter not found then record it
                            no_formatters.append(destination['format'])
                            continue

                        formatter.set_destination(destination, subscriber)
                        formatted_docs = formatter.format(self.filter_document(doc),
                                                          subscriber,
                                                          subscriber_codes.get(subscriber[config.ID_FIELD]))

                        for idx, publish_data in enumerate(formatted_docs):
                            if not isinstance(publish_data, dict):
                                pub_seq_num, formatted_doc = publish_data
                                formatted_docs[idx] = {'published_seq_num': pub_seq_num,
                                                       'formatted_item': formatted_doc}
                            else:
                                assert 'published_seq_num' in publish_data and 'formatted_item' in publish_data, \
                                    "missing keys in publish_data"

                        for publish_queue_item in formatted_docs:
                            publish_queue_item['item_id'] = doc['item_id']
                            publish_queue_item['item_version'] = doc[config.VERSION]
                            publish_queue_item['subscriber_id'] = subscriber[config.ID_FIELD]
                            publish_queue_item['codes'] = subscriber_codes.get(subscriber[config.ID_FIELD])
                            publish_queue_item['destination'] = destination
                            # publish_schedule is just to indicate in the queue item is create via scheduled item
                            publish_queue_item[PUBLISH_SCHEDULE] = get_utc_schedule(doc, PUBLISH_SCHEDULE) or None
                            publish_queue_item['unique_name'] = doc.get('unique_name', None)
                            publish_queue_item['content_type'] = doc.get('type', None)
                            publish_queue_item['headline'] = doc.get('headline', None)
                            publish_queue_item['publishing_action'] = self.published_state
                            publish_queue_item['ingest_provider'] = \
                                ObjectId(doc.get('ingest_provider')) if doc.get('ingest_provider') else None
                            publish_queue_item['associated_items'] = associations.get(subscriber[config.ID_FIELD], [])
                            publish_queue_item['priority'] = subscriber.get('priority')

                            if doc.get(PUBLISHED_IN_PACKAGE):
                                publish_queue_item[PUBLISHED_IN_PACKAGE] = doc[PUBLISHED_IN_PACKAGE]
                            try:
                                encoded_item = publish_queue_item.pop('encoded_item')
                            except KeyError:
                                pass
                            else:
                                binary = io.BytesIO(encoded_item)
                                publish_queue_item['encoded_item_id'] = app.storage.put(binary)
                            publish_queue_item.pop(ITEM_STATE, None)

                            # content api delivery will be marked as SUCCESS in queue
                            get_resource_service('publish_queue').post([publish_queue_item])

                            queued = True
                except Exception:
                    logger.exception("Failed to queue item for id {} with headline {} for subscriber {}."
                                     .format(doc.get(config.ID_FIELD), doc.get('headline'), subscriber.get('name')))

            return no_formatters, queued
        except Exception:
            raise

    def _embed_package_items(self, package):
        """Embeds all package items in the package document."""
        for group in package.get(GROUPS, []):
            if group[GROUP_ID] == ROOT_GROUP:
                continue
            for ref in group[REFS]:
                if RESIDREF not in ref:
                    continue
                package_item = get_resource_service('published').find_one(req=None, item_id=ref[RESIDREF],
                                                                          _current_version=ref[config.VERSION])
                if not package_item:
                    msg = _('Can not find package {package} published item {item}').format(
                        package=package['item_id'], item=ref['residRef'])
                    raise SuperdeskPublishError(500, msg)
                package_item[config.ID_FIELD] = package_item['item_id']
                ref['package_item'] = package_item
        return package

    def _get_subscribers_for_package_item(self, package_item):
        """Finds the list of subscribers for a given item in a package

        :param package_item: item in a package
        :return list: List of subscribers
        """
        query = {'$and': [{'item_id': package_item[config.ID_FIELD]},
                          {'publishing_action': package_item[ITEM_STATE]}]}

        return self._get_subscribers_for_previously_sent_items(query)

    def _get_subscribers_for_previously_sent_items(self, lookup):
        """Returns list of subscribers that have previously received the item.

        :param dict lookup: elastic query to filter the publish queue
        :return: list of subscribers and list of product codes per subscriber
        """
        req = ParsedRequest()
        subscribers = []
        subscriber_codes = {}
        associations = {}
        queued_items = list(get_resource_service('publish_queue').get(req=req, lookup=lookup))

        if len(queued_items) > 0:
            subscriber_ids = {}
            for queue_item in queued_items:
                subscriber_id = queue_item['subscriber_id']
                if not subscriber_ids.get(subscriber_id):
                    subscriber_ids[subscriber_id] = False
                    if queue_item.get('destination', {}).get('delivery_type') == 'content_api':
                        subscriber_ids[subscriber_id] = True

                subscriber_codes[subscriber_id] = queue_item.get('codes', [])
                if queue_item.get('associated_items'):
                    associations[subscriber_id] = list(set(associations.get(subscriber_id, [])) |
                                                       set(queue_item.get('associated_items', [])))

            query = {'$and': [{config.ID_FIELD: {'$in': list(subscriber_ids.keys())}}]}
            subscribers = list(get_resource_service('subscribers').get(req=None, lookup=query))
            for s in subscribers:
                s['api_enabled'] = subscriber_ids.get(s.get(config.ID_FIELD))

        return subscribers, subscriber_codes, associations

    def filter_subscribers(self, doc, subscribers, target_media_type):
        """Filter subscribers to whom the current document is going to be delivered.

        :param doc: Document to publish/kill/correct
        :param subscribers: List of Subscribers that might potentially get this document
        :param target_media_type: Valid values are - Wire, Digital.
        :return: List of of filtered subscribers and list of product codes per subscriber.
        """
        filtered_subscribers = []
        subscriber_codes = {}
        existing_products = {p[config.ID_FIELD]: p for p in
                             list(get_resource_service('products').get(req=None, lookup=None))}
        global_filters = deepcopy([gf['cf'] for gf in self.filters.get('content_filters', {}).values() if
                                   gf['cf'].get('is_global', True)])

        # apply global filters
        self.conforms_global_filter(global_filters, doc)

        for subscriber in subscribers:
            if target_media_type and subscriber.get('subscriber_type', '') != SUBSCRIBER_TYPES.ALL:
                can_send_digital = subscriber['subscriber_type'] == SUBSCRIBER_TYPES.DIGITAL
                if target_media_type == SUBSCRIBER_TYPES.WIRE and can_send_digital or \
                        target_media_type == SUBSCRIBER_TYPES.DIGITAL and not can_send_digital:
                    continue

            conforms, skip_filters = self.conforms_subscriber_targets(subscriber, doc)
            if not conforms:
                continue

            if not self.conforms_subscriber_global_filter(subscriber, global_filters):
                continue

            product_codes = self._get_codes(subscriber)
            subscriber_added = False
            subscriber['api_enabled'] = False
            # validate against direct products
            result, codes = self._validate_article_for_subscriber(doc,
                                                                  subscriber.get('products'),
                                                                  existing_products)
            if result:
                product_codes.extend(codes)
                if not subscriber_added:
                    filtered_subscribers.append(subscriber)
                    subscriber_added = True

            if content_api.is_enabled():
                # validate against api products
                result, codes = self._validate_article_for_subscriber(doc,
                                                                      subscriber.get('api_products'),
                                                                      existing_products)
                if result:
                    product_codes.extend(codes)
                    subscriber['api_enabled'] = True
                    if not subscriber_added:
                        filtered_subscribers.append(subscriber)
                        subscriber_added = True

            if skip_filters and not subscriber_added:
                # if targeted subscriber and has api products then send it to api.
                if subscriber.get('api_products'):
                    subscriber['api_enabled'] = True
                filtered_subscribers.append(subscriber)
                subscriber_added = True

            # unify the list of codes by removing duplicates
            if subscriber_added:
                subscriber_codes[subscriber[config.ID_FIELD]] = list(set(product_codes))

        return filtered_subscribers, subscriber_codes

    def _validate_article_for_subscriber(self, doc, products, existing_products):
        """Validate the article for subscriber

        :param dict doc: Document to be validated
        :param list products: list of product ids
        :param dict existing_products: Product lookup
        :return tuple bool, list: Boolean flag to add subscriber or not and list of product codes.
        """
        add_subscriber, product_codes = False, []

        if not products:
            return add_subscriber, product_codes

        for product_id in products:
            # check if the product filter conforms with the story
            product = existing_products.get(product_id)

            if not product:
                continue

            if not self.conforms_product_targets(product, doc):
                continue

            if self.conforms_content_filter(product, doc):
                # gather the codes of products
                product_codes.extend(self._get_codes(product))
                add_subscriber = True

        return add_subscriber, product_codes

    def _filter_subscribers_for_associations(self, subscribers, doc, target_media_type, existing_associations):
        """Filter the subscriber for associations.

        :param list subscribers: list of subscriber that are going to receive parent item.
        :param dict doc: item with associations
        :param dict existing_associations: existing associations
        :param target_media_type: Valid values are - Wire, Digital.
        """
        associations = {}

        if not doc.get(ASSOCIATIONS) or not subscribers:
            return associations

        for assoc, item in doc.get(ASSOCIATIONS).items():
            if not item:
                continue

            assoc_subscribers = set()
            assoc_id = item.get(config.ID_FIELD)
            filtered_subscribers, subscriber_codes = self.filter_subscribers(item,
                                                                             deepcopy(subscribers),
                                                                             target_media_type)

            for subscriber in filtered_subscribers:
                # for the validated subscribers
                subscriber_id = subscriber.get(config.ID_FIELD)
                if not associations.get(subscriber_id):
                    associations[subscriber_id] = []

                associations[subscriber_id].append(assoc_id)
                assoc_subscribers.add(subscriber_id)

            for subscriber_id, items in existing_associations.items():
                # for the not validated associated item but previously published to the subscriber.
                if assoc_id in items and assoc_id not in associations.get(subscriber_id, []):
                    if not associations.get(subscriber_id):
                        associations[subscriber_id] = []

                    associations[subscriber_id].append(assoc_id)
                    assoc_subscribers.add(subscriber_id)

            item['subscribers'] = list(assoc_subscribers)

        return associations

    def _update_associations(self, original, updates):
        """Update the associations

        :param dict original: original item
        :param dict updates: updates item
        """
        if not updates:
            return

        for subscriber, items in updates.items():
            if items:
                original[subscriber] = list(set(original.get(subscriber, [])) | set(updates.get(subscriber, [])))

    def conforms_product_targets(self, product, article):
        """Check product targets.

        Checks if the given article has any target information and if it does
        it checks if the product satisfies any of the target information

        :param product: Product to test
        :param article: article
        :return:
            bool: True if the article conforms the targets for the given product
        """
        geo_restrictions = product.get('geo_restrictions')

        # If not targeted at all then Return true
        if not BasePublishService().is_targeted(article, 'target_regions'):
            return geo_restrictions is None

        if geo_restrictions:
            for region in article.get('target_regions', []):
                if region['qcode'] == geo_restrictions and region['allow']:
                    return True
                if region['qcode'] != geo_restrictions and not region['allow']:
                    return True
        return False

    def conforms_subscriber_targets(self, subscriber, article):
        """Check subscriber targets.

        Checks if the given article has any target information and if it does
        it checks if the subscriber satisfies any of the target information

        :param subscriber: Subscriber to test
        :param article: article
        :return:
            bool: True/False if the article conforms the targets
            bool: True if the given subscriber is specifically targeted, False otherwise
        """
        # If not targeted at all then Return true
        if not BasePublishService().is_targeted(article, 'target_subscribers') and \
                not BasePublishService().is_targeted(article, 'target_types'):
            return True, False

        subscriber_type = subscriber.get('subscriber_type')

        for t in article.get('target_subscribers', []):
            if str(t.get('_id')) == str(subscriber['_id']):
                return True, True

        if subscriber_type:
            for t in article.get('target_types', []):
                if t['qcode'] == subscriber_type and t['allow']:
                    return True, False
                if t['qcode'] != subscriber_type and not t['allow']:
                    return True, False

        # If there's a region target then continue with the subscriber to check products
        if BasePublishService().is_targeted(article, 'target_regions'):
            return True, False

        # Nothing matches so this subscriber doesn't conform
        return False, False

    def conforms_content_filter(self, product, doc):
        """Checks if the document matches the subscriber filter

        :param product: Product where the filter is used
        :param doc: Document to test the filter against
        :return:
        True if there's no filter
        True if matches and permitting
        False if matches and blocking
        False if doesn't match and permitting
        True if doesn't match and blocking
        """
        content_filter = product.get('content_filter')

        if content_filter is None or 'filter_id' not in content_filter or content_filter['filter_id'] is None:
            return True

        service = get_resource_service('content_filters')
        filter = self.filters.get('content_filters', {}).get(content_filter['filter_id'], {}).get('cf')
        does_match = service.does_match(filter, doc, self.filters)

        if does_match:
            return content_filter['filter_type'] == 'permitting'
        else:
            return content_filter['filter_type'] == 'blocking'

    def conforms_global_filter(self, global_filters, doc):
        """Check global filter

        Checks if document matches the global filter

        :param global_filters: List of all global filters
        :param doc: Document to test the global filter against
        """
        service = get_resource_service('content_filters')
        for global_filter in global_filters:
            global_filter['does_match'] = service.does_match(global_filter, doc, self.filters)

    def conforms_subscriber_global_filter(self, subscriber, global_filters):
        """Check global filter for subscriber

        Checks if subscriber has a override rule against each of the
        global filter and if not checks if document matches the global filter

        :param subscriber: Subscriber to get if the global filter is overriden
        :param global_filters: List of all global filters
        :return: True if at least one global filter is not overriden
        and it matches the document
        False if global filter matches the document or all of them overriden
        """

        gfs = subscriber.get('global_filters', {})
        for global_filter in global_filters:
            if gfs.get(str(global_filter[config.ID_FIELD]), True):
                # Global filter applies to this subscriber
                if global_filter.get('does_match'):
                    return False
        return True

    def _extend_subscriber_items(self, subscriber_items, subscribers, item, package_item_id, subscriber_codes):
        """Extends the subscriber_items with the given list of subscribers for the item

        :param subscriber_items: The existing list of subscribers
        :param subscribers: New subscribers that item has been published to - to be added
        :param item: item that has been published
        :param package_item_id: package_item_id
        """
        item_id = item[config.ID_FIELD]
        for subscriber in subscribers:
            sid = subscriber[config.ID_FIELD]
            item_list = subscriber_items.get(sid, {}).get('items', {})
            item_list[item_id] = package_item_id
            subscriber_items[sid] = {'subscriber': subscriber,
                                     'items': item_list,
                                     'codes': subscriber_codes.get(sid, [])}

    def _get_codes(self, item):
        if item.get('codes'):
            return [c.strip() for c in item.get('codes').split(',') if c]
        else:
            return []

    @staticmethod
    def filter_document(doc):
        """
        Filter document:
        1. Remove fields that should not be there given it's profile.
        2. Remove `None` valued renditions.

        :param dict doc: document to filter
        :return: dict filtered document
        """

        # remove fields that should not be there given it's profile.
        doc = apply_schema(doc)

        # remove `None` valued renditions.
        for association_key in doc.get(ASSOCIATIONS, {}):
            association = doc[ASSOCIATIONS][association_key]
            if not association:
                continue

            renditions = association.get('renditions', {})
            for null_rendition_key in [k for k in renditions if not renditions[k]]:
                del doc[ASSOCIATIONS][association_key]['renditions'][null_rendition_key]

        return doc




import superdesk

from flask import request, current_app as app, json
from superdesk.resource import build_custom_hateoas

from apps.archive.archive import ArchiveResource, ArchiveService
from apps.archive.common import CUSTOM_HATEOAS


def elastic_filter():
    where = request.args.get('where')
    assert where, 'where not set'

    params = json.loads(where)
    guid = params['guid']
    uri = params.get('uri') or guid

    query = {
        'bool': {
            'should': [
                {'term': {'refs.uri': uri}},
                {'term': {'refs._id': guid}},
                {'term': {'refs.guid': guid}},
            ],
        },
    }

    LINKS_HOURS = app.config.get('LINKS_MAX_HOURS')
    if LINKS_HOURS:
        query['bool'].update({
            'minimum_should_match': 1,
            'must': {
                'range': {
                    'versioncreated': {
                        'gte': 'now-{}h'.format(int(LINKS_HOURS)),
                    },
                },
            },
        })

    return query


class LinksResource(ArchiveResource):
    item_methods = []
    resource_methods = ['GET']
    datasource = ArchiveResource.datasource.copy()
    datasource.update({
        'source': 'archive',
        'elastic_filter_callback': elastic_filter,
        'elastic_filter': {'bool': {
            'must_not': {'term': {'version': 0}}
        }},
    })


class LinksService(ArchiveService):

    def enhance_items(self, items):
        super().enhance_items(items)
        for item in items:
            build_custom_hateoas(CUSTOM_HATEOAS, item)

    def get(self, req, lookup):
        req.where = None  # it's handled in the elastic_filter
        return super().get(req, lookup)


def init_app(_app):
    superdesk.register_resource('links', LinksResource, LinksService, _app=_app)



"""
News resource
=============

It is an alias for archive without filtering out published items.
"""


from superdesk.resource import build_custom_hateoas
from apps.archive.archive import ArchiveResource, ArchiveService
from apps.archive.common import CUSTOM_HATEOAS


class NewsResource(ArchiveResource):
    datasource = ArchiveResource.datasource.copy()
    datasource.update({
        'source': 'archive',
        'elastic_filter': {'bool': {
            'must_not': {'term': {'version': 0}}
        }},
    })

    resource_methods = ['GET']
    item_methods = []


class NewsService(ArchiveService):

    def enhance_items(self, items):
        super().enhance_items(items)
        for item in items:
            build_custom_hateoas(CUSTOM_HATEOAS, item)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import copy
from superdesk import get_resource_service, config, app
from superdesk.resource import Resource


default_allowed_filters = ['anpa_category',
                           'urgency',
                           'keywords',
                           'priority',
                           'slugline',
                           'type',
                           'source',
                           'headline',
                           'ednote',
                           'body_html',
                           'genre',
                           'subject',
                           'desk',
                           'stage',
                           'sms',
                           'place',
                           'ingest_provider',
                           'embargo',
                           'featuremedia',
                           'anpa_take_key',
                           'agendas']


class FilterConditionResource(Resource):
    schema = {
        'name': {
            'type': 'string',
            'required': True,
            'nullable': False,
            'empty': False,
            'iunique': True
        },
        'field': {
            'type': 'string',
            'empty': False,
            'nullable': False,
            'allowed': default_allowed_filters,
        },
        'operator': {
            'type': 'string',
            'allowed': ['in',
                        'nin',
                        'like',
                        'notlike',
                        'startswith',
                        'endswith',
                        'match',
                        'eq',
                        'ne',
                        'lt',
                        'lte',
                        'gt',
                        'gte',
                        'exists'],
            'empty': False,
            'nullable': False,
        },
        'value': {
            'type': 'string',
            'empty': False,
            'nullable': False,
        }
    }

    additional_lookup = {
        'url': 'regex("[\w,.:-]+")',
        'field': 'name'
    }

    privileges = {'POST': 'content_filters',
                  'PATCH': 'content_filters',
                  'DELETE': 'content_filters'}

    mongo_indexes = {
        'name_1': ([('name', 1)], {'unique': True}),
    }

    def pre_request_post(self, request):
        self._init_allowed_filters()

    def pre_request_patch(self, request, lookup):
        self._init_allowed_filters()

    def _init_allowed_filters(self):
        self.schema['field']['allowed'] = copy.copy(default_allowed_filters)
        self.schema['field']['allowed'].extend(app.config.get('EXCLUDED_VOCABULARY_FIELDS', []))
        lookup = {'_id': {'$nin': self.schema['field']['allowed']}, 'type': 'manageable'}
        for vocabulary in get_resource_service('vocabularies').get(req=None, lookup=lookup):
            self.schema['field']['allowed'].append(vocabulary[config.ID_FIELD])



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging
import superdesk

from flask import current_app as app
from superdesk.activity import add_activity, ACTIVITY_UPDATE
from superdesk.services import BaseService
from superdesk import get_resource_service
from superdesk.errors import SuperdeskApiError
from superdesk.notification import push_notification
from superdesk.utils import compare_preferences
from superdesk.resource import Resource


logger = logging.getLogger(__name__)


class RolesResource(Resource):
    schema = {
        'name': {
            'type': 'string',
            'required': True,
            'nullable': False,
            'empty': False,
            'iunique': True
        },
        'description': {
            'type': 'string'
        },
        'privileges': {
            'type': 'dict',
            'schema': {},
            'allow_unknown': True,
        },
        'is_default': {
            'type': 'boolean'
        },

        #: author role to be automatically set when user creates an item
        #: .. versionadded:: 1.29
        'author_role': {
            'type': 'string',
            'nullable': True,
        },

        #: author role to be automatically set when user edits an item
        #: .. versionadded:: 1.29
        'editor_role': {
            'type': 'string',
            'nullable': True,
        },
    }

    privileges = {'POST': 'roles', 'DELETE': 'roles', 'PATCH': 'roles'}

    mongo_indexes = {
        'name_1': ([('name', 1)], {'unique': True}),
    }


class RolesService(BaseService):

    def on_update(self, updates, original):
        if updates.get('is_default'):
            # if we are updating the role that is already default that is OK
            if original.get('is_default'):
                return
            self.remove_old_default()

    def on_create(self, docs):
        for doc in docs:
            # if this new one is default need to remove the old default
            if doc.get('is_default'):
                self.remove_old_default()

    def on_delete(self, docs):
        if docs.get('is_default'):
            raise SuperdeskApiError.forbiddenError('Cannot delete the default role')
        # check if there are any users in the role
        user = get_resource_service('users').find_one(req=None, role=docs.get('_id'))
        if user:
            raise SuperdeskApiError.forbiddenError('Cannot delete the role, it still has users in it!')

    def remove_old_default(self):
        # see if there is already a default role and set it to no longer default
        role_id = self.get_default_role_id()
        # make it no longer default
        if role_id:
            role = self.find_one(req=None, is_default=True)
            get_resource_service('roles').update(role_id, {"is_default": False}, role)

    def get_default_role_id(self):
        role = self.find_one(req=None, is_default=True)
        return role.get('_id') if role is not None else None

    def on_updated(self, updates, role):
        self.__send_notification(updates, role)

    def __send_notification(self, updates, role):
        role_id = role['_id']

        role_users = superdesk.get_resource_service('users').get_users_by_role(role_id)
        notified_users = [user['_id'] for user in role_users]

        if 'privileges' in updates:
            privileges_updated = False
            added, removed, modified = compare_preferences(role.get('privileges', {}), updates['privileges'])
            if len(removed) > 0 or (1, 0) in modified.values():
                push_notification('role_privileges_revoked', updated=1, role_id=str(role_id))
                privileges_updated = True
            if len(added) > 0 or (0, 1) in modified.values():
                activity = add_activity(ACTIVITY_UPDATE,
                                        'role {{role}} has been granted new privileges: Please re-login.',
                                        self.datasource,
                                        notify=notified_users,
                                        can_push_notification=False,
                                        role=role.get('name'))
                push_notification('activity', _dest=activity['recipients'])
                privileges_updated = True
            if privileges_updated:
                app.on_role_privileges_updated(role, role_users)

        else:
            push_notification('role', updated=1, user_id=str(role_id))



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import os
from unittest import mock
from unittest.mock import MagicMock

from apps.auth.errors import CredentialsAuthError
from superdesk.errors import SuperdeskApiError
from superdesk.tests import TestCase
from superdesk import get_resource_service
from apps.ldap.commands import ImportUserProfileFromADCommand


if os.environ.get('LDAP_SERVER', ''):

    def get_mock_connection():
        """Create a mock ldap connection object.

        :return {object}: mock ldap connection object.
        """
        connection = MagicMock()
        connection.search(return_value=True)
        connection.response = [
            {
                'attributes': {
                    'sn': ['Bar'],
                    'givenName': ['Foo'],
                    'displayName': ['Foo Bar'],
                    'ipPhone': '+1234567890',
                    'email': 'foo@bar.com'
                }
            }
        ]

        return connection

    @mock.patch('apps.ldap.ldap.Connection', return_value=get_mock_connection())
    class ImportUsersTestCase(TestCase):
        """for testing import user using ldap."""

        def test_import_user_using_command(self, mock_ldap_connection):
            user = {'username': 'user', 'password': 'pwd', 'user_to_import': 'barf'}
            cmd = ImportUserProfileFromADCommand()

            cmd.run(user['username'], user['password'], user['user_to_import'])
            auth_user = get_resource_service('auth_db').authenticate({'username': 'barf', 'password': 'dummy'})
            self.assertEqual(auth_user['username'], user['user_to_import'])

            cmd.run(user['username'], user['password'], 'BARF')
            auth_user2 = get_resource_service('auth_db').authenticate({'username': 'barf', 'password': 'dummy'})
            self.assertEqual(auth_user2['username'], user['user_to_import'])

            self.assertEqual(auth_user2['_id'], auth_user['_id'])

        @mock.patch('apps.ldap.ldap.get_user', return_value={'username': 'user'})
        def test_import_profile_for_already_imported_user_raises_exception(self, mock_ldap_connection, mock_get_user):
            with self.assertRaises(SuperdeskApiError) as context:
                service = get_resource_service('import_profile')
                doc = {'username': 'user', 'password': 'pwd', 'profile_to_import': 'barf'}
                users = service.post([doc])
                self.assertIsNotNone(users)
                self.assertEqual(len(users), 1)
                doc = {'username': 'user', 'password': 'pwd', 'profile_to_import': 'BARF '}
                service.post([doc])

            ex = context.exception
            self.assertEqual(ex.message, 'User already exists in the system.')
            self.assertEqual(ex.status_code, 400)
            self.assertDictEqual(ex.payload, {'profile_to_import': 1})

        @mock.patch('apps.ldap.ldap.get_user', return_value={'username': 'user1'})
        def test_import_user_by_not_logged_in_user_raises_exception(self, mock_ldap_connection, mock_get_user):
            with self.assertRaises(SuperdeskApiError) as context:
                service = get_resource_service('import_profile')
                doc = {'username': 'user', 'password': 'pwd', 'profile_to_import': 'barf'}
                service.post([doc])

            ex = context.exception
            self.assertEqual(ex.message, 'Invalid Credentials.')
            self.assertEqual(ex.status_code, 403)
            self.assertDictEqual(ex.payload, {'credentials': 1})

        @mock.patch('apps.ldap.ldap.ADAuth.authenticate_and_fetch_profile',
                    side_effect=CredentialsAuthError(credentials={'username': 'test'}, error='test'))
        @mock.patch('apps.ldap.ldap.get_user', return_value={'username': 'user'})
        def test_user_import_profile_with_invalid_credentials(self, mock_ldap_connection, mock_auth, mock_get_user):
            with self.assertRaises(SuperdeskApiError) as context:
                service = get_resource_service('import_profile')
                doc = {'username': 'user', 'password': 'pwd', 'profile_to_import': 'barf'}
                service.post([doc])

            ex = context.exception
            self.assertEqual(ex.message, 'Invalid Credentials.')
            self.assertEqual(ex.status_code, 403)
            self.assertDictEqual(ex.payload, {'credentials': 1})



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2019 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import operator
from functools import reduce
import superdesk
from flask import current_app as app


class ProdApiService(superdesk.Service):
    """
    Base service for production api services
    """

    # specify fields to exclude from a response
    # NOTE: it's possible to specify nested key as `update._links.self.title`
    # in this case only item['update']['_links']['self']['title'] will be removed.
    excluded_fields = {
        '_etag',
        '_type',
        '_updated',
        '_created',
        '_current_version',
        '_links',
        'fields_meta',
    }

    def on_fetched(self, result):
        """
        Event handler when a collection of items is retrieved from database.
        Post-process a fetched items.
        :param dict result: dictionary contaning the list of fetched items and
         some metadata, e.g. pagination info.
        """
        for doc in result['_items']:
            self._process_fetched_object(doc)

    def on_fetched_item(self, doc):
        """
        Event handler when a single item is retrieved from a database.
        Post-process a fetched item.
        :param dict docs: fetched items from a database.
        """
        self._process_fetched_object(doc)

    def _process_fetched_object(self, doc):
        """
        Does some processing on the document fetched from database.
        :param dict document: MongoDB document to process
        """
        self._remove_excluded_fields(doc)
        self._process_item_renditions(doc)

    def _remove_excluded_fields(self, item):
        """
        Remove keys from an item
        """
        for key in self.excluded_fields:
            keys = key.split('.')
            if keys[:-1]:
                # pop last key only from `item`
                try:
                    reduce(operator.getitem, keys[:-1], item).pop(keys[-1], None)
                except KeyError:
                    pass
            item.pop(key, None)

    def _process_item_renditions(self, doc):
        """
        Replace `href` in `renditions` and in `body_html`
        """

        def _process(item):
            for _k, v in item.get('renditions', {}).items():
                if v and 'media' in v:
                    media = v.pop('media')
                    old_href = v.get('href')
                    new_href = app.media.url_for_media(media, v.get('mimetype'))
                    v['href'] = new_href
                    # replace href in body
                    if old_href and doc.get('body_html'):
                        # no need to use lxml here, it will be to much
                        doc['body_html'] = doc['body_html'].replace(old_href, new_href)

        if doc.get('renditions'):
            _process(doc)

        for v in [v for v in doc.get('associations', {}).values() if v]:
            _process(v)
            self._remove_excluded_fields(v)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import os
from superdesk.tests import TestCase
from superdesk.io.feed_parsers.ap_media import APMediaFeedParser
import json


class APMediaTestCase(TestCase):
    vocab = [{'_id': 'genre', 'items': [{'name': 'Current'}]}]

    def setUp(self):
        with self.app.app_context():
            self.app.data.insert('vocabularies', self.vocab)
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.normpath(os.path.join(dirname, '../fixtures', self.filename))
        provider = {'name': 'Test'}
        with open(fixture) as fp:
            self.item = APMediaFeedParser().parse(json.load(fp), provider)


class SimpleTestCase(APMediaTestCase):

    filename = 'ap_media_picture.json'

    def test_headline(self):
        self.assertEqual(self.item.get('headline'), 'US Paul Simon Poets Society')
